#include <stdlib.h>

#include "httpd.h"
#include "http_config.h"
//#include "http_protocol.h" /* printing to the request */
#include "libpq-fe.h"

#include "co_lib.h"
#include "co_log.h"
#include "co_action.h"
#include "co_array.h"
#include "co_string.h"
#include "co_defines.h"

#define SQLBUFSIZE 10000
#define JSBUFSIZE 10000
#define MAXSELECTPAR 1000

#define CO_ACTION_ID 0
#define CO_ACTION_OPERATION 1
#define CO_ACTION_MAIN_CLASS_NAME 2
#define CO_ACTION_CLASS_NAME1 3
#define CO_ACTION_OBJECT_ID1 4
#define CO_ACTION_CLASS_NAME2 5
#define CO_ACTION_OBJECT_ID2 6
#define CO_ACTION_CLASS_NAME3 7
#define CO_ACTION_OBJECT_ID3 8
#define CO_ACTION_CLASS_NAME4 9
#define CO_ACTION_OBJECT_ID4 10
#define CO_ACTION_PROPERTY_NAMES_AND_VALUES 11

static const char *operation[] = {
    "noop",
    "update",
    "insert",
    "delete",
    "undelete",
    "create"
};

static const char *client_operation[] = {
    "co_noop",
    "co_refresh",
    "co_insert",
    "co_delete",
    "co_undelete",
    "co_create"
};

static char SQLString[SQLBUFSIZE];

/* exported method: adds action to the queue, "spits" local javascript */
/* ASSUME: class_names, prop_name do not require js-encoding */
void
cou_queueActionX( co_context *cnx, 
	const char *main_class_name,
	const char **class_names,
	const int obj_ids[], /* who? */
	int operation, /* PropertyNameAndValueList */ ...)
{
    va_list ap; // for argument list

    PGresult *res;
    int i, j;
    int begin_array;
    const char *prop_name; // for va_arg
    const char *prop_value; // for va_arg

    va_start(ap, operation);

    if(cnx->putJS) {
	request_rec *r = cnx->request;

	cou_jsprintf(cnx, "var o=new Array(); o[0]=new Object(); o[0].local=1; o[0].class_name=\'%s\';\n", main_class_name);
	cou_jsputs( cnx, "o[0].d=new Object();");
	cou_jsputs( cnx, "o[0].classes = new Array();");
	i = 0;
	while (class_names[i])
	{
	 if (strcmp(class_names[i],"")) 
	 { 
	  cou_jsprintf(cnx, "o[0].%s_id=%d; ",class_names[i],obj_ids[i]);
	  cou_jsprintf(cnx, "o[0].classes[%d]={name:\"%s\",id:%d}; ", i, class_names[i], obj_ids[i]);
	 }
	 i++;
	}

	do
	{
	    prop_name = va_arg(ap,char *);
	    if (prop_name) prop_value = (char *) cou_buildEncodedJS(cnx, va_arg(ap,char *));
	    else prop_value = NULL;
	    if (prop_value)
	    {
		cou_jsprintf(cnx, "o[0].d.%s=\'%s\';\n", prop_name, prop_value);
	    }
	    else prop_name = NULL;
	}
	while (prop_name);
	
	cou_jsprintf(cnx, "%s(o);\n", client_operation[operation]);
    };


    if ((operation<0) || (operation>CO_OP_MAX)) cou_error( cnx, "Bad operation id in cou_queueAction");

    sprintf(SQLString, "
    SELECT co_%s_action_insert(%d,%d,'%s',
    ", cnx->par_area, cnx->id_session, operation, main_class_name);
    i = 0;
    for (j=0;j<4;j++)
	if (class_names[i])
	{
	    sprintf(SQLString,"%s'%s',%d,", SQLString, class_names[i], obj_ids[i]);
	    i++;
	}
	else sprintf(SQLString,"%s'',0,",SQLString);
    
//    sprintf(SQLString,"%s'%s','%s'", SQLString, prop_name, value);

    va_start(ap, operation);
    sprintf(SQLString,"%s'{",SQLString); // array values
    
    begin_array = 1;
    do 
    {
	prop_name = va_arg(ap,char *);
	if (prop_name) prop_value = (char *) cou_buildEncodedSQLArrayField(cnx, va_arg(ap,char *) );
	else prop_value = NULL;
	if (prop_value)
	{
	    if (!begin_array) sprintf( SQLString,"%s,\"%s\",\"%s\"", SQLString, prop_name, prop_value);
            else { sprintf( SQLString,"%s\"%s\",\"%s\"", SQLString, prop_name, prop_value); begin_array = 0; }
	}
	else prop_name = NULL;
    }
    while (prop_name);

    sprintf(SQLString,"%s}')",SQLString);

    cou_logError(cnx,CO_LOG_DEBUG,"%s",SQLString);

    res = PQexec(cnx->dbconn, SQLString);
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
         cou_error( cnx, "Error while queuing action" );
    if (atoi(PQgetvalue(res, 0, 0))) cou_error( cnx, "Bad return value while queuing action" );

    PQclear(res);
    va_end(ap);
}

void
cou_queueActionXX( co_context *cnx,
	const char *main_class_name,
	const char **class_names,
	const int obj_ids[], /* who? */
	int operation, const char** p_v /* PropertyNameAndValueList */ )
{
    PGresult *res;
    int i, j;
    int begin_array;
    int iter;
    const char *prop_name; // for va_arg
    const char *prop_value; // for va_arg
    
    iter = 0;

    if(cnx->putJS) {
	request_rec *r = cnx->request;

	cou_jsprintf(cnx, "var o=new Array(); o[0]=new Object(); o[0].local=1; o[0].class_name=\'%s\';\n", main_class_name);
	cou_jsputs(cnx, "o[0].d=new Object();");
	cou_jsputs(cnx, "o[0].classes = new Array();");
	
	i = 0;
	while (class_names[i])
	{
	 if (strcmp(class_names[i],"")) 
	 {
	  cou_jsprintf(cnx, "o[0].%s_id=%d; ",class_names[i],obj_ids[i]);
	  cou_jsprintf(cnx, "o[0].classes[%d]={name:\"%s\",id:%d}; ", i, class_names[i], obj_ids[i]);
	 }
	 i++;
	}

	
	do 
	{
	    prop_name = (char *) p_v[iter++];
	    if (prop_name) prop_value = (char *) cou_buildEncodedJS(cnx, p_v[iter++]); 
	    else prop_value = NULL;
	    if (prop_value)
	    {
		cou_jsprintf(cnx, "o[0].d.%s=\'%s\';\n", prop_name, prop_value);
	    }
	    else prop_name = NULL;
	}
	while (prop_name);
	
	cou_jsprintf(cnx, "%s(o);\n", client_operation[operation]);
    };


    if ((operation<0) || (operation>CO_OP_MAX)) cou_error( cnx, "Bad operation id in cou_queueAction");

    sprintf(SQLString, "
    SELECT co_%s_action_insert(%d,%d,'%s',
    ", cnx->par_area, cnx->id_session, operation, main_class_name);
    i = 0;
    for (j=0;j<4;j++)
    if (class_names[i])
    {
     sprintf(SQLString,"%s'%s',%d,", SQLString, class_names[i], obj_ids[i]);
     i++;
    }
    else sprintf(SQLString,"%s'',0,",SQLString);
    
//    sprintf(SQLString,"%s'%s','%s'", SQLString, prop_name, value);

    iter = 0;

    sprintf(SQLString,"%s'{",SQLString); // array values
    
    begin_array = 1;
    do 
    {
	prop_name = (char *) p_v[iter++];
	if (prop_name) prop_value = (char *) cou_buildEncodedSQLArrayField(cnx, p_v[iter++] );
	else prop_value = NULL;
	if (prop_value)
	{
	    if (!begin_array) sprintf( SQLString,"%s,\"%s\",\"%s\"", SQLString, prop_name, prop_value);
	    else { sprintf( SQLString,"%s\"%s\",\"%s\"", SQLString, prop_name, prop_value); begin_array = 0; }
	}
	else prop_name = NULL;
    }
    while (prop_name);
    
    sprintf(SQLString,"%s}')",SQLString);

    cou_logError(cnx,CO_LOG_DEBUG,"%s",SQLString);

    res = PQexec(cnx->dbconn, SQLString);
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
         cou_error( cnx, "Error while queuing action" );
    if (atoi(PQgetvalue(res, 0, 0))) cou_error( cnx, "Bad return value while queuing action" );

    PQclear(res);
}



/* private module methods */

void
cox_sendWaitingActionsX( co_context *cnx )
{
    char SQLString[SQLBUFSIZE];
    PGresult *res;
    int nFields, nTuples;
    int i,j;
    co_array_of_pointers * geta_classes;
    co_array_of_pointers * geta_ids = NULL;
    const char* par_class;
    int *par_id_ptr=NULL;
    int par_id=1;
    int err=0;
    int geta_err=0;
    int ok;
    int in_dq; // in double-quoted string
    int prev_bs; // previous character was backslash
    int NClasses, NIds; // length counters

    const char* class_names[] = { NULL, NULL, NULL, NULL, NULL};
    int obj_ids[] = { 0, 0, 0, 0, 0};
    
    char *p_nvs; // property names and values
    int last_action_index;

    geta_classes = (co_array_of_pointers *) cou_getParamArray(cnx,"geta_classes","=",MAXSELECTPAR,&geta_err);
    cou_logError(cnx,CO_LOG_DEBUG,"action2 geta_classes error= %d",geta_err);
    if (geta_err==0) /* geta_classes found */
    {
     geta_ids = (co_array_of_pointers *)cou_getParamArray(cnx,"geta_ids","=",MAXSELECTPAR,&geta_err);
     cou_logError(cnx,CO_LOG_DEBUG,"action2 geta_ids error= %d",geta_err);
     if (geta_err==0) 
     {
      i = 0;
      NClasses = coa_GetElementCount(cnx, geta_classes);
      NIds = coa_GetElementCount(cnx, geta_ids);
      if (NIds >= NClasses)
        while (i < NClasses)
        {
         if (strcmp(coa_GetElement(cnx,geta_ids,i),"")) obj_ids[i] = atoi(coa_GetElement(cnx,geta_ids,i));
          // no more geta_ids
         i++;
        }
     }
     else
     {
	geta_err = 0;
     }
     par_class = NULL; /* not used when we have geta_classes */
    }
    else /* geta_classes not found */
    {
     par_class = (const char*)cou_getParam(cnx,"par_class",CO_STRING_SQL_NAME, &err);
     par_id_ptr = (int*)cou_getParam(cnx,"par_id",CO_INTEGER, &err);
     if (par_id_ptr) par_id = *par_id_ptr;
    }
    
    if (snprintf(SQLString, SQLBUFSIZE,
"
SELECT 
ACT.id,
ACT.operation,
(SELECT class_name FROM co_%s_meta_class WHERE id=ACT.main_class_id),
(SELECT class_name FROM co_%s_meta_class WHERE id=ACT.class_id1),
ACT.object_id1,
(SELECT class_name FROM co_%s_meta_class WHERE id=ACT.class_id2),
ACT.object_id2,
(SELECT class_name FROM co_%s_meta_class WHERE id=ACT.class_id3),
ACT.object_id3,
(SELECT class_name FROM co_%s_meta_class WHERE id=ACT.class_id4),
ACT.object_id4,
ACT.property_names_and_values
FROM co_%s_action AS ACT
WHERE
    ACT.session_id <> %d AND    
    ACT.id > (SELECT last_action_index FROM co_%s_session SES WHERE SES.id=%d)",
cnx->par_area, cnx->par_area, cnx->par_area, cnx->par_area, cnx->par_area,
cnx->par_area, cnx->id_session, cnx->par_area, cnx->id_session)
>= SQLBUFSIZE) cou_error( cnx, "SQL command buffer too small in cou_queueAction" );

if (!(geta_err==0)) /* geta_err<0 */
{
 sprintf(SQLString,"%s
    AND ACT.class_id1 = (SELECT id FROM co_%s_meta_class WHERE class_name='%s')
    AND ACT.object_id1 = %d
    ORDER by ACT.id;
 ",SQLString, cnx->par_area, par_class, par_id);
}
else
{
 i = 0;
 NClasses = coa_GetElementCount(cnx,geta_classes);
 if (geta_ids) {
    NIds = coa_GetElementCount(cnx, geta_ids);
 }
 else {
    NIds = 0; /* getParamArray did not found ids (or there was an error) */
 }
 while (i < NClasses)
 {
  if (strcmp(coa_GetElement(cnx,geta_classes,i),"")) {
    sprintf(SQLString,"%s
	AND ACT.class_id%d = (SELECT id FROM co_%s_meta_class WHERE class_name='%s')",SQLString, i+1, cnx->par_area, (const char *) coa_GetElement(cnx,geta_classes,i));
    if (i < NIds) sprintf(SQLString,"%s
	AND ACT.object_id%d = %d",SQLString, i+1, obj_ids[i]);
  }
  i++;
 }
 sprintf(SQLString,"%s\n ORDER by ACT.id;",SQLString);
}

    cou_logError( cnx, CO_LOG_DEBUG, SQLString);
    
    res = PQexec(cnx->dbconn, SQLString);
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
         cou_error( cnx, "Error while getting actions (waiting)" );
    
    nFields = PQnfields(res);
    
    nTuples = PQntuples(res);

    if (nTuples==0) { /* nothing to do */
	PQclear(res);
	return;
    }
    
    if (cnx->putJS)
    {
     request_rec *r = cnx->request;

     for (i = 0; i < nTuples; i++)
     {
        p_nvs = PQgetvalue(res, i, CO_ACTION_PROPERTY_NAMES_AND_VALUES);

 	j = 3; // should be constant
	ok = 0;
	while (j>=0)
	{
	 class_names[j] = PQgetvalue(res, i, CO_ACTION_CLASS_NAME1+(2*j));
	 if (strcmp(class_names[j],"")) 
	 {
	  obj_ids[j] =atoi(PQgetvalue(res, i, CO_ACTION_CLASS_NAME1+(2*j)+1));
	  ok = 1;
	 }
	 else if (!ok) class_names[j] = NULL;
         j--;
	}

	cou_jsprintf(cnx, "var o=new Array(); o[0]=new Object(); o[0].classes=new Array();");
	cou_jsprintf(cnx, "o[0].class_name=\'%s\';\n", PQgetvalue(res, i, CO_ACTION_MAIN_CLASS_NAME));

	j = 0;
	while (class_names[j])
	{
	 if (strcmp(class_names[j],"")) 
	 {
	  cou_jsprintf(cnx, "o[0].%s_id=%d; ",class_names[j],obj_ids[j]);
	  cou_jsprintf(cnx, "o[0].classes[%d]={name:\"%s\",id:%d}; ", j, class_names[j], obj_ids[j]);
	 }
	 j++;
	}

	if (strcmp(p_nvs,"{}"))
	{
	 j = 0;
	 ok = 1; in_dq = 0; prev_bs = 0;
	 while (p_nvs[j])
	 {
	  if ((!prev_bs) && (p_nvs[j]=='"')) in_dq = 1-in_dq;
	  if ((!in_dq) && (p_nvs[j]==','))
	  {
	   if (ok) p_nvs[j] = ':';
	   ok = 1-ok;
	  }
	  if (p_nvs[j]==92) prev_bs = 1; else prev_bs = 0;
	  j++;
	 }
	 //new_value_js_encoded = (char*)ap_palloc( cnx->request->pool, JSBUFSIZE*sizeof(char) );
	 //cou_JSEncode( PQgetvalue(res, i, CO_ACTION_VALUE) ,new_value_js_encoded);
	 //ap_rprintf( r, "o[0].d.%s=\'%s\';\n", property_name, new_value_js_encoded);
	 cou_jsprintf(cnx, "o[0].d = %s;\n",p_nvs);
	}

	cou_jsprintf(cnx, "%s(o);\n", client_operation[atoi(PQgetvalue(res, i, CO_ACTION_OPERATION ))]);

    }
   }

    last_action_index = atoi(PQgetvalue(res, nTuples-1, CO_ACTION_ID ));
    cou_logError( cnx, CO_LOG_DEBUG, "last_action_index=%d",last_action_index);
    PQclear(res);

    sprintf(SQLString, 
"
UPDATE co_%s_session
SET last_action_index = %d
WHERE id = %d;
",
cnx->par_area, last_action_index, cnx->id_session );
cou_logError( cnx, CO_LOG_DEBUG, SQLString );

    res = PQexec(cnx->dbconn, SQLString);
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
         cou_error( cnx, "Error while updating last_action_index" );
    PQclear(res);
}

void
cox_sendLocalActions( co_context *cnx )
{
//    cou_putJS( cnx, "alert('co_action_get in action: local actions');" );
}
