#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

#include "co_lib.h"
#include "co_defines.h"
#include "co_mod.h"

#include "co_array.h"
#include "mod_coffice.h"

struct co_array_of_pointers
{
   int magic;
   void ** elems;
   int NElements;
   int MAXElemens;

   int type;
};

#undef __SEND_MESG_DEBUG
#define __SEND_MESG_DEBUG(x...)

#define ARRAY_FUN_COMMON(fun_name) \
  char MESGString[MESGSIZE]; \
  pool * p = cnx->request->pool; \
  __SEND_MESG_DEBUG("entering co_array_of_pointers::" # fun_name) \
  __ASSERT(coa_IsArray(cnx,coar), "ERROR: co_array_of_pointers::" # fun_name " trying to execute for incorect coar") \

#define ARRAY_FUN_COMMON_NULL(fun_name) \
  char MESGString[MESGSIZE]; \
  pool * p = cnx->request->pool; \
  __SEND_MESG_DEBUG("entering co_array_of_pointers::" # fun_name) \
  __ASSERT_NULL(coa_IsArray(cnx,coar), "ERROR: co_array_of_pointers::" # fun_name " trying to execute for incorect coar") \

#define ARRAY_FUN_COMMON2(fun_name) \
  char MESGString[MESGSIZE]; \
  pool * p = cnx->request->pool; \
  __SEND_MESG_DEBUG("co_array_of_pointers::" # fun_name)


// Constructor
co_array_of_pointers * coa_Create(co_context * cnx)
{
  co_array_of_pointers * ret;
  ARRAY_FUN_COMMON2("coa_Create")

  ret = ap_pcalloc(p, sizeof(co_array_of_pointers)); // fills with zeros
  ret->magic = __ARRAY_MAGIC_NUMBER;
  ret->elems = ap_pcalloc(p, __ARRAY_BLOCK_SIZE*sizeof(void *));
  ret->MAXElemens = __ARRAY_BLOCK_SIZE;
  ret->type = COARRAY_TYPE_STRING;
  return ret;
}

co_array_of_pointers * coa_CreateTyped(co_context * cnx, int type)
{
  co_array_of_pointers * ret; 
  ARRAY_FUN_COMMON2("coa_Create")

  ret = ap_pcalloc(p, sizeof(co_array_of_pointers));
  ret->magic = __ARRAY_MAGIC_NUMBER;
  ret->elems = ap_pcalloc(p, __ARRAY_BLOCK_SIZE*sizeof(void *));
  ret->MAXElemens = __ARRAY_BLOCK_SIZE;
  ret->type = type;
  return ret;
}


//////////////////////////////////////////////////////////////////////
//TEST METHODS

int coa_IsArray(co_context * cnx, const co_array_of_pointers * coar)
{
   if(coar==NULL) return 0;
   if(coar->magic != __ARRAY_MAGIC_NUMBER) return 0;
   return 1;
} 


// Get Methods
int   coa_GetElementCount(co_context * cnx, const co_array_of_pointers * coar)
{
  ARRAY_FUN_COMMON("coa_GetElementCount")
  return coar->NElements;
}

void * coa_GetElement(co_context * cnx, const co_array_of_pointers * coar, int elem_num)
{
  ARRAY_FUN_COMMON_NULL("coa_GetElement")

  __ASSERT_NULL( (elem_num >= 0) , "coa_GetElement: elem_num < 0, elem_num: %d", elem_num)
  __ASSERT_NULL( (elem_num < coa_GetElementCount(cnx, coar)) , "coa_GetElement: elem_num >= nelts, elem_num: %d", elem_num)

  return coar->elems[elem_num];
}

const char * coa_GetStringElement(co_context * cnx, const co_array_of_pointers * coar, int elem_num)
{
  ARRAY_FUN_COMMON_NULL("coa_GetStringElement")

  __ASSERT_NULL( coar->type == COARRAY_TYPE_STRING , "coa_GetElement: wrong type of array, should be COARRAY_TYPE_STRING but is: %d", coar->type)
  __ASSERT_NULL( (elem_num >= 0) , "coa_GetElement: elem_num < 0, elem_num: %d", elem_num)
  __ASSERT_NULL( (elem_num < coa_GetElementCount(cnx, coar)) , "coa_GetElement: elem_num >= nelts, elem_num: %d", elem_num)

  return coar->elems[elem_num];
}

char * coa_GetCopyOfElement(co_context * cnx, const co_array_of_pointers * coar, int elem_num)
{
  ARRAY_FUN_COMMON_NULL("coa_GetElement")

  __ASSERT_NULL( coar->type == COARRAY_TYPE_STRING , "coa_GetElement: wrong type of array, should be COARRAY_TYPE_STRING but is: %d", coar->type)
  __ASSERT_NULL( (elem_num >= 0) , "coa_GetElement: elem_num < 0, elem_num: %d", elem_num)
  __ASSERT_NULL( (elem_num < coa_GetElementCount(cnx, coar)) , "coa_GetElement: elem_num >= nelts, elem_num: %d", elem_num)

  return ap_pstrdup(p, coar->elems[elem_num]);
}


// Manipulate Methods
int   coa_AppendElement(co_context * cnx, co_array_of_pointers * coar, void * new_value)
{
  ARRAY_FUN_COMMON("coa_AppendElement")

  if(coar->NElements+1>= coar->MAXElemens)
  {
    void * old_elems = coar->elems;
    coar->MAXElemens += __ARRAY_BLOCK_SIZE;
    coar->elems = ap_pcalloc(cnx->request->pool, coar->MAXElemens*sizeof(void*));
    memcpy(coar->elems, old_elems, (coar->MAXElemens - __ARRAY_BLOCK_SIZE) * sizeof(void *) );
  }
  coar->elems[coar->NElements]=new_value;
  coar->NElements++;
  return coar->NElements;
}

int   coa_AppendCopyOfElement(co_context * cnx, co_array_of_pointers * coar, const char * new_value)
{
  ARRAY_FUN_COMMON("coa_AppendElement")

  __ASSERT( coar->type == COARRAY_TYPE_STRING , "coa_AppendCopyOfElement: wrong type of array, should be COARRAY_TYPE_STRING but is: %d", coar->type)

  if(coar->NElements+1>= coar->MAXElemens)
  {
    void * old_elems = coar->elems;
    coar->MAXElemens += __ARRAY_BLOCK_SIZE;
    coar->elems = ap_pcalloc(cnx->request->pool, coar->MAXElemens*sizeof(void*));
    memcpy(coar->elems, old_elems, (coar->MAXElemens - __ARRAY_BLOCK_SIZE) * sizeof(void *) );
  }
  coar->elems[coar->NElements] = ap_pstrdup(p, new_value);
  coar->NElements++;
  return coar->NElements;
}


int coa_AppendSplit(co_context * cnx, co_array_of_pointers * coar, const char * str_to_split, const char * delim)
{
   const char * curr = str_to_split;
   const char * next = str_to_split;
   int i = 0;

   ARRAY_FUN_COMMON("coa_AppendSplit")
  
   __ASSERT(str_to_split!=NULL, "in coa_AppendSplit string to split is NULL")
   __ASSERT(delim!=NULL, "in coa_AppendSplit delimiter is NULL")

   while((next=strstr(curr, delim)))
   {
      char * to_append =  ap_pstrndup(p, curr, next - curr);
      __SEND_MESG_DEBUG("APPENDING: %s", to_append); 
      coa_AppendElement(cnx, coar, to_append);
      curr = next + strlen(delim);
      i++;
   }
    __SEND_MESG_DEBUG("APPENDING LAST: %s", curr);
    coa_AppendElement(cnx, coar, (void *) curr);

    return i+1;
}  

int   coa_SetElement(co_context * cnx, co_array_of_pointers * coar, int elem_num, void * new_value)
{
   ARRAY_FUN_COMMON("coa_SetElement")

  __ASSERT( (elem_num >= 0) , "coa_SetElement: elem_num < 0, elem_num: %d", elem_num)
  __ASSERT( (elem_num < coa_GetElementCount(cnx, coar)) , "coa_SetElement: elem_num >= nelts, elem_num: %d", elem_num)
  
  coar->elems[elem_num]=new_value;
  return elem_num;
}

// Copy Methods

co_array_of_pointers * coa_Copy(co_context * cnx, const co_array_of_pointers * coar)
{

  co_array_of_pointers * ret;
  int NElements;
  int i;
  ARRAY_FUN_COMMON_NULL("coa_Copy")

  __ASSERT_NULL( coar->type == COARRAY_TYPE_STRING , "coa_Copy: wrong type of array, should be COARRAY_TYPE_STRING but is: %d", coar->type)

  ret = coa_Create(cnx);
  NElements = coa_GetElementCount(cnx, coar);
  for(i=0; i<NElements; i++)
  {
    coa_AppendElement(cnx, ret, coa_GetCopyOfElement(cnx, coar, i));
  }
  return ret;
}



