#define SERVER_EVENT_COUNT 2

typedef enum {
    SERVER_EVENT_ONBEFORE = 0,
    SERVER_EVENT_ONAFTER = 1
} co_server_event;

typedef struct method_info
{
	const char *area_name;
	const char *source_class_name;
	const char *method_name;
	const char *target_class_name;
	int id_script[SERVER_EVENT_COUNT]; // js_code ids for all server_events
	int id_method[SERVER_EVENT_COUNT]; // meta_method ids for all server_events
	int profiled; // whether method call should be profiled
	int sec_type; //type of security (object == 0, class == 1)

	// this fields are used during security checking
	// they are equall to values passed in params or
	// if not present in params we use values from url
	const char *check_par_area;
	const char *check_par_class;
	int check_par_id;

	// 1 if acl checking passed succesfully
	int acl;
} method_info;

typedef struct req_params_values
{
	const char *par_area;
	const char *par_class;
	int par_id;
} req_params_values;

typedef struct req_params_control
{
	int par_area;
	int par_class;
	int par_id;
} req_params_control;


typedef struct req_params
{
	int magic_par_area;
	int magic_par_class;
	int magic_par_id;

	req_params_control * control;
	req_params_values * values;
} req_params;




extern
const char **
cou_getContainerInfo( 
    co_context *ctx, 
    const char* area_name, 
    const char* class_name);

extern
void
cou_broker( 
    const char* area_name, const char* class_name, const char* method_name,
    co_context *ctx, void *in_params );

void
cox_executeEventScript(
    co_context *ctx, 
    method_info *m_i, 
    co_server_event s_e, 
    void *in_params, int *err);

void
cox_executeEventMethod(
    co_context *ctx, 
    method_info *m_i, 
    co_server_event s_e, int *err);

extern
const char *
cox_getTargetClassName( 
    co_context *ctx, 
    const char* area_name, const char* class_name, const char* method_name, 
    int *err);

void
cox_gatherMethodInfo(
    co_context *ctx,
    method_info *m_i,
    int *err);
