#include "httpd.h"
/* PostgreSQL access functions */
#include "libpq-fe.h"

#include "co_jsinheritance.h"

#define CO_INH_CLASS 0
#define CO_INH_METHOD 1
#define CO_INH_TARGET 2

static const char *query_classes =
"
select class_name
from co_sys_meta_vtable
 union
select class_name
from co_usr_meta_vtable
 union
select class_name
from co_sec_meta_vtable
 group by class_name
  order by class_name
  ;
";

static const char *query_implementations =
"
select class_name, method_name, target_class_name from co_sys_meta_vtable
 where class_name != target_class_name
 union
select class_name, method_name, target_class_name from co_usr_meta_vtable
 where class_name != target_class_name
 union
select class_name, method_name, target_class_name from co_sec_meta_vtable
 where class_name != target_class_name
 group by class_name, method_name, target_class_name
  order by class_name, method_name ;
";

// set when server starts
static const char *scripts[2] = { NULL, NULL };
#define SCRIPT_LINE_SIZE 512

static
const char *
cop_buildClassesScript( pool *p, PGconn *conn )
{
    PGresult *pgres;
    int ntuples;
    int i;
    char *_script;
    const char *script;
    const char *class_name;
    
    pgres = PQexec(conn, query_classes);
    if (PQresultStatus(pgres) != PGRES_TUPLES_OK)
    {
        PQclear(pgres);
        return NULL;
    }

    ntuples = PQntuples(pgres);
    _script = (char*)ap_palloc( p, ntuples*SCRIPT_LINE_SIZE*2);
    if (!_script) return NULL;

    script = _script;
    for (i=0; i<ntuples; i++) {
	class_name = PQgetvalue(pgres, i, CO_INH_CLASS);
	_script += snprintf( _script, SCRIPT_LINE_SIZE, 
	    /* "function %s(arg_area) { this.init = COObject_init; this.init(arg_area,'%s'); };\n",*/
	    "function class_%s(arg_area) { this.strClass='%s'; this.strArea = arg_area; };\n",
	    class_name, class_name
	);
    }

    PQclear(pgres);
    return script;
}

static
const char *
cop_buildImplementationsScript( pool *p, PGconn *conn )
{
    PGresult *pgres;
    int ntuples;
    int i;
    char *_script;
    const char *script;
    const char *class_name;
    const char *method_name;
    const char *target_name;
    const char *js_method_name;
    
    pgres = PQexec(conn, query_implementations);
    if (PQresultStatus(pgres) != PGRES_TUPLES_OK)
    {
        PQclear(pgres);
        return NULL;
    }

    ntuples = PQntuples(pgres);
    _script = (char*)ap_palloc( p, ntuples*SCRIPT_LINE_SIZE*3);
    if (!_script) return NULL;

    script = _script;
    for (i=0; i<ntuples; i++) {
	class_name = PQgetvalue(pgres, i, CO_INH_CLASS);
	method_name = PQgetvalue(pgres, i, CO_INH_METHOD);
	target_name = PQgetvalue(pgres, i, CO_INH_TARGET);
	if (strcmp(method_name, "delete")==0) js_method_name="del";
	else js_method_name = method_name;

	_script += snprintf( _script, SCRIPT_LINE_SIZE, 
	    "class_%s.prototype.RAW_%s = class_%s_RAW_%s;\n"
	    "class_%s.prototype.EXT_%s = class_%s_EXT_%s;\n"
	    "class_%s.prototype.%s = class_%s_%s;\n", 
	    class_name, js_method_name, target_name, method_name,
	    class_name, js_method_name, target_name, method_name,
	    class_name, js_method_name, target_name, method_name
	);
    }

    PQclear(pgres);
    return script;
}

// class.prototype.method = target_method
const char **
cox_buildInheritanceScript( pool *p, PGconn *conn )
{
    scripts[0] = cop_buildClassesScript( p, conn );
    scripts[1] = cop_buildImplementationsScript( p, conn );
    return scripts;
}