#ifndef CORECORDSET
#define CORECORDSET

#include "httpd.h"
#include "libpq-fe.h"
#include "co_array.h"

#define COREC_MAGIC_NUMBER    1234567

#define COREC_EMPTY           -1
#define COREC_OLD_TYPE         0
#define COREC_EXTENDED_TYPE    1
#define COREC_ARRAY_TYPE       2

typedef struct co_recordset co_recordset;

// CREATORS
co_recordset * cor_CreateEmpty       (co_context * cnx);
co_recordset * cor_CreateEmptyNamed  (co_context * cnx, const char * name);
co_recordset * cor_CreateOnPGResult  (co_context * cnx, PGresult * res);
co_recordset * cor_CreateOnSQLString (co_context * cnx, const char * SQLString);

// TEST METHODS
int      cor_IsRecordset        (co_context * cnx, const co_recordset * corec);
int      cor_IsEmpty            (co_context * cnx, const co_recordset * corec);
int      cor_IsCleared          (co_context * cnx, const co_recordset * corec);

// GET METHODS
int            cor_GetTupleCount         (co_context * cnx, const co_recordset * corec);
int            cor_GetFieldCount         (co_context * cnx, const co_recordset * corec);
int            cor_GetFieldNum           (co_context * cnx, const co_recordset * corec, const char * field_name);
const char *   cor_GetFieldName          (co_context * cnx, const co_recordset * corec, int field_num);
const char *   cor_GetValue              (co_context * cnx, const co_recordset * corec, int tuple_num, int field_num);
const char *   cor_GetValueByName        (co_context * cnx, const co_recordset * corec, int tuple_num, const char * field_name);
char *         cor_GetCopyOfValueByName  (co_context * cnx, const co_recordset * corec, int tuple_num, const char * field_name);
char *         cor_GetCopyOfValue        (co_context * cnx, const co_recordset * corec, int tuple_num, int field_num);
char *         cor_GetADOXMLSerialization(co_context * cnx, const co_recordset * corec);
co_array_of_pointers * cor_GetCopyOfColumnByName(co_context * cnx, const co_recordset * corec, const char* field_name);


co_array_of_pointers * cor_GetCopyOfRecord(co_context * cnx, const co_recordset * corec, int tuple_num);

// MANIPULATE METHODS
int      cor_AppendRowset        (co_context * cnx, co_recordset * corec, PGresult * res);
int      cor_AppendOnSQLString   (co_context * cnx, co_recordset * corec, const char * SQLString);
int      cor_AppendRecord        (co_context * cnx, co_recordset * corec, const co_recordset * source, int tuple_num);
int      cor_AppendRecordset     (co_context * cnx, co_recordset * corec, const co_recordset * source);
int      cor_SetFieldName        (co_context * cnx, co_recordset * corec, int field_num, char * new_name);


// CLEAR METHODS
int      cor_Clear                 (co_context * cnx, co_recordset * corec);

// JS CREATION METHODS
int cou_createJSFromRecordset      (co_context *ctx, const co_recordset * corec, const char* callback); 
int cou_createJSHeadFromRecordset  (co_context *ctx, const co_recordset * corec);
int cou_createJSBodyFromRecordset  (co_context *ctx, const co_recordset * corec);
int cou_createJSCallback           (co_context *ctx, const char * callback);


#endif
