#include "postgres.h"
#include "executor/spi.h"
#include "utils/elog.h"
#include <stdio.h>
#define BUFSIZE 100000
#define SQLBUFSIZE 10000
#define MAXROW 100000
char *
textout_lh(text *vlena)
{
	int			len;
	char	   *result;

	if (vlena == NULL)
	{
		result = (char *) palloc(2);
		result[0] = '-';
		result[1] = '\0';
		return result;
	}
	len = VARSIZE(vlena) - VARHDRSZ;
	//elog(NOTICE, "textout, size: %d, \n",len);			      
	result = (char *) palloc(len + 1);
	memmove(result, VARDATA(vlena), len);
	result[len] = '\0';

#ifdef CYR_RECODE
	convertstr(result, len, 1);
#endif

	return result;
}


//#define t(x) textout(x)
#define t(x) textout_lh(x)
//#define t(x) x->vl_dat

void set_text_size(text *ptext, int32 new_text_size)
{
    VARATT_SIZEP(ptext) = VARATT_SIZEP(ptext) & VARATT_MASK_FLAGS ;
    VARATT_SIZEP(ptext) = VARATT_SIZEP(ptext) | new_text_size;
    //elog(NOTICE, "set_text_size:  %d, %d \n",new_text_size, VARSIZE(ptext) );			      
}                 

text *
c_text(text *arg1, text *arg2)
{
    //elog(NOTICE, "c_text: %s, %s \n",t(arg1), t(arg2));			      
    int32 new_text_size = VARSIZE(arg1) + VARSIZE(arg2) - VARHDRSZ;
    text *new_text = (text *) palloc(new_text_size);
    //elog(NOTICE, "c_text: %s, %s \n",t(arg1), t(arg2));			      
    memset((void *) new_text, 0, new_text_size);
    //VARSIZE(new_text) = new_text_size;
    set_text_size(new_text, new_text_size);
    strncpy(VARDATA(new_text), VARDATA(arg1), VARSIZE(arg1)-VARHDRSZ);
    strncat(VARDATA(new_text), VARDATA(arg2), VARSIZE(arg2)-VARHDRSZ);
    return(new_text);
}   

text *
cr_text(char *arg)
{
    int32 new_text_size= strlen(arg)+VARHDRSZ;
    //elog(NOTICE, "cr_text: %s, %d \n",arg, new_text_size);			      
    text *new_text = (text *) palloc(new_text_size);
    //elog(NOTICE, "cr_text2:\n");			      
    memset((void *) new_text, 0, new_text_size);
    //elog(NOTICE, "cr_text3:\n");			      
    //VARSIZE(new_text) = new_text_size;
    set_text_size(new_text, new_text_size);
    strncpy(VARDATA(new_text), arg, new_text_size);
    //elog(NOTICE, "cr_text4:\n");			      
    return(new_text);
}   


text *
getpath(text *area, text *class_name, text *column_name, text *delimiter, int id_root, int id_item)
{
    int ret;
    int proc = 0;
    int max;

    text *new_text = cr_text("");
    text *text2 = cr_text("");
    text *text3 = cr_text("");
    int len = 0;
    char sql[SQLBUFSIZE];
    char id_buf[BUFSIZE];
    char id_tmp[100];
    int id_last = 0;

    memset((void *) id_buf, 0, BUFSIZE);

    SPI_connect();
    
    do{
      snprintf(sql, SQLBUFSIZE, "SELECT id_parent, %s FROM co_%s_%s WHERE id = %d", 
                                      t(column_name),t(area), t(class_name), id_item);
	//elog(NOTICE, "getpath: %s \n",sql);			      
      ret = SPI_exec(sql, MAXROW);
      proc = SPI_processed;
	//elog(NOTICE, "getpath2: \n");
      if ( ret == SPI_OK_SELECT && SPI_processed > 0 )
      {
          TupleDesc tupdesc = SPI_tuptable->tupdesc;
          SPITupleTable *tuptable = SPI_tuptable;
          HeapTuple tuple = tuptable->vals[0];
       
          /* getting values */
          id_last = id_item;
	  //elog(NOTICE, "getpath2a: \n");
          id_item = atoi(SPI_getvalue(tuple, tupdesc, 1));
	  //elog(NOTICE, "getpath2b: \n");
	//  text2 = cr_text(SPI_getvalue(tuple, tupdesc, 2));
	  //elog(NOTICE, "getpath2c: \n");
	  //c_text(text2,delimiter);
	  //elog(NOTICE, "getpath2d: \n");
	  new_text = c_text(c_text(cr_text(SPI_getvalue(tuple, tupdesc, 2)),delimiter),new_text);
	 // elog(NOTICE, "getpath2e: \n");

          snprintf(id_tmp,100,"[%d]",id_item);
          if(strstr(id_buf,id_tmp)>0)
          {
             elog(NOTICE, "LOOPS in getpath on area: %s class: %s id: %d", t(area), t(class_name), id_item);
             len = strlen(VARDATA(new_text));
	     new_text = c_text(new_text, cr_text("***LOOPS***"));
             break;
          }
          else
          {
             strncat(id_buf,id_tmp,BUFSIZE-strlen(id_buf));
          };
       }
       else
       {
         break;
       }
    }
    while(id_last!=id_root && id_item != 0 && id_last!=0);
    //elog(NOTICE, "getpath3: \n");
    SPI_finish();
    //elog(NOTICE, "getpath4: \n");
    return (new_text);
}                    

text *
getpathx(text *area, text *class_name, text *id_column, text * foreign_column, text *column_name, text *delimiter, text *text_id_root, text *text_id_item)
{
    int ret;
    int proc = 0;
    int max;

    text *new_text = cr_text("");
    int len = 0;
    char sql[SQLBUFSIZE];
    char id_buf[BUFSIZE];
    char id_tmp[100];
    text *text_id_last;

    memset((void *) id_buf, 0, BUFSIZE);

    SPI_connect();
    
    do{
      snprintf(sql, SQLBUFSIZE, "SELECT %s, %s FROM co_%s_%s WHERE %s = '%s'", 
                          t(foreign_column), t(column_name),t(area), t(class_name), t(id_column), t(text_id_item));
      ret = SPI_exec(sql, MAXROW);
      proc = SPI_processed;

      if ( ret == SPI_OK_SELECT && SPI_processed > 0 )
      {
          TupleDesc tupdesc = SPI_tuptable->tupdesc;
          SPITupleTable *tuptable = SPI_tuptable;
          HeapTuple tuple = tuptable->vals[0];
       
          /* getting values */
          text_id_last = cr_text(t(text_id_item));
          text_id_item = cr_text(SPI_getvalue(tuple, tupdesc, 1));

	  new_text = c_text(c_text(cr_text(SPI_getvalue(tuple, tupdesc, 2)),delimiter),new_text);

          snprintf(id_tmp,100,"[%s]",t(text_id_item));
          if(strstr(id_buf,id_tmp)>0)
          {
             elog(NOTICE, "LOOPS in getpath on area: %s class: %s id: %s", t(area), t(class_name), t(text_id_item));
             len = strlen(VARDATA(new_text));
	     new_text = c_text(new_text, cr_text("***LOOPS***"));
             break;
          }
          else
          {
             strncat(id_buf,id_tmp,BUFSIZE-strlen(id_buf));
          };
       }
       else
       {
         break;
       }
    }
    while(strcmp(t(text_id_last),t(text_id_root))!=0 && strcmp(t(text_id_item),"0")!=0 && strcmp(t(text_id_last),"0")!=0 && strlen(t(text_id_item))>0);

    SPI_finish();
    return (new_text);
}                    



/* this is not for use by clients! */

text *
u_gettree(text *area, text *class_name, text *column_name, text *delimiter, text *in, int id_item)
{
    int ret;
    int proc = 0;
    int max;
    int err;
    text *ret_text;
    int len = 0;
    char sql[SQLBUFSIZE];

    elog (NOTICE, "GETTREE id: %d  in: %s", id_item, t(in));

    // getting all children
    snprintf(sql, SQLBUFSIZE, "SELECT id, %s FROM co_%s_%s WHERE id_parent = %d", 
                                      t(column_name),t(area), t(class_name), id_item);

    elog (NOTICE, "GETTREE sql: %s", sql);
    ret = SPI_exec(sql, MAXROW);
    proc = SPI_processed;

    if ( ret == SPI_OK_SELECT && SPI_processed > 0 )
    {
      TupleDesc tupdesc = SPI_tuptable->tupdesc;
      SPITupleTable *tuptable = SPI_tuptable;

      ret_text = cr_text("");

      for(ret = 0; ret <proc; ret++)
      {
         HeapTuple tuple = tuptable->vals[ret];
	 // getting new text
         text *new_text_prop = cr_text(SPI_getvalue(tuple, tupdesc, 2));

         // adding in+delimiter+text
         text *new_text = c_text(c_text(in, delimiter), new_text_prop);
         
         // getting values 
         id_item = atoi(SPI_getvalue(tuple, tupdesc, 1));
	 
         elog (NOTICE, "GETTREE MAMY NOWY %s", t(new_text));        
         // recurence
         ret_text = c_text(ret_text, u_gettree(area,class_name,column_name,delimiter, new_text, id_item));
      }; // for each tuple

      return ret_text;
    }
    //if nothing found we return what we have been given
    return(in);
}   

/* this is used by clients */
text *
gettree(text *area, text *class_name, text *column_name, text *delimiter, text *in, int id_item)
{
    text *ret;
    SPI_connect();
    ret = u_gettree(area,class_name,column_name,delimiter,in,id_item);
    SPI_finish();
    return ret;
};    


text * 
recure(text *in, int a)
{
   text *ret_text;
   text *new_text;
   int ret, proc;
   int new_text_size;
   char sql[SQLBUFSIZE];

   snprintf(sql, SQLBUFSIZE, "SELECT class_name FROM co_sys_meta_class WHERE id = %d",a); 

   SPI_connect();
   if(a<=0) return in;

   ret = SPI_exec(sql, MAXROW);
   proc = SPI_processed;

   if ( ret == SPI_OK_SELECT && SPI_processed > 0 )
   {
      TupleDesc tupdesc = SPI_tuptable->tupdesc;
      SPITupleTable *tuptable = SPI_tuptable;
      HeapTuple tuple = tuptable->vals[0];

     /* getting values */
     new_text= cr_text(SPI_getvalue(tuple, tupdesc, 1));
     ret_text = (text *) c_text(c_text(in, new_text),recure(in, a-1));
   }
   else
   {
     ret_text = (text *) c_text(in,recure(in, a-1));
   };
   SPI_finish();

   return ret_text;
}                    

text *
execq(text *sql, int cnt)
{
    int ret;
    int proc = 0;
    int32 new_text_size = 10000;    
    text *new_text = (text *) palloc(new_text_size);
    int len = 0;

    SPI_connect();
    ret = SPI_exec(t(sql), cnt);
    proc = SPI_processed;


    memset((void *) new_text, 0, new_text_size);


    /*
    * If this is SELECT and some tuple(s) fetched -
    * returns tuples to the caller via elog (NOTICE).
   */
 
   if ( ret == SPI_OK_SELECT && SPI_processed > 0 )
   {
       TupleDesc tupdesc = SPI_tuptable->tupdesc;
       SPITupleTable *tuptable = SPI_tuptable;
        char buf[8192];
        int i;
        for (ret = 0; ret < proc; ret++)
        {
            HeapTuple tuple = tuptable->vals[ret];
 
            for (i = 1, buf[0] = 0; i <= tupdesc->natts; i++)
            {
                sprintf(buf + strlen (buf), " %s%s",
                    SPI_getvalue(tuple, tupdesc, i),
                    (i == tupdesc->natts) ? " " : " |");
            }
           len = strlen(VARDATA(new_text));
           snprintf(VARDATA(new_text)+len, new_text_size-len, "wiersz: %s %d\n", buf, len);
        }
    }
    //VARSIZE(new_text) =  new_text_size; 
    set_text_size(new_text, new_text_size);
    SPI_finish();
    return (new_text);
}

