#include "types.t"
#include "defines.t"

// CLASS CLIPBOARD

class clipboard

{
  area "sys";
  area "sec";
  area "usr";

  inherits generic_acl;

#include "check_acl.t"

  container_class '';
  container_column_name '';

 // PROPERTIES

#include "generic_acl_props.t"

  prop user_id
  {
    type CO_INTEGER;
    default "0";
    params '';
    representant 1;
    foreign '';
    foreign_prop '';
    position 200;
  };

  prop object_serialization
  {
    type CO_STRING;
    default "''";
    params '';
    representant 1;
    foreign '';
    foreign_prop '';
    position 300;
  };

  prop owner_area
  {
    type CO_STRING;
    default "''";
    params '';
    representant 1;
    foreign '';
    foreign_prop '';
    position 400;
  };

  prop owner_class
  {
    type CO_STRING;
    default "''";
    params '';
    representant 1;
    foreign '';
    foreign_prop '';
    position 500;
  };

// METHODS

	method insertClipboardObject
	{
		doc "Puts object of some type in to current users clipboard slot. Each user has only one slot per area!";
		id_permission __PERMISSION_UPDATE;
		sec_type __METHOD_SEC_TYPE_CLASS;
    
		__PARAMS_HEADER

		param serialization 
		{ 
			doc "Some value that we want to put into clipboard";
			type CO_STRING; 
		};

		param owner_area
		{
			doc "Area of the object that we put into clipboard";
			type CO_STRING;
		};

		param owner_class
		{
			doc "Class name of the object that we put into clipboard";
			type CO_STRING;
		};
    
		declare language_c
		{
			C "int user_id;";
		};
    
		body language_c
		{
			C
			%%
			user_id = cnx->id_user;
      
			if (!serialization) serialization = "";
      
			__CREATE_QUERY("
				UPDATE co_%s_clipboard SET object_serialization='%s', owner_area='%s', owner_class='%s' WHERE user_id=%d",
				cnx->par_area, serialization, owner_area, owner_class, user_id
			)
      
			__SQL_EXEC
			%%
			;
		};
	}; // method

	method retrieveClipboardObject
	{
		id_permission __PERMISSION_GET;
		sec_type __METHOD_SEC_TYPE_CLASS;

		doc "Retrieves serialization of the object in current area and for current user.";

		__PARAMS_HEADER
		param owner_area
		{
			doc "Area of the object that we are getting from clipboard";
			type CO_STRING;
		};

		param owner_class
		{
			doc "Class name of the object  that we are getting from clipboard";
			type CO_STRING;
		};
    
		declare language_c
		{
			C "int user_id;";
		};

		body language_c
		{
			C
			%%
			user_id = cnx->id_user;
      
			__CREATE_QUERY("
				SELECT object_serialization FROM co_%s_clipboard 
				WHERE user_id=%d AND owner_area='%s' AND owner_class='%s'
				ORDER BY id DESC",
				cnx->par_area, user_id, owner_area, owner_class
			)

			__SQL_GET_RESULT

			if (ok) {
				cou_setReturnValue(cnx,cor_GetCopyOfValue(cnx, corec, 0,0),CO_STRING);
			} else {
				cou_setReturnValue(cnx,"",CO_STRING);
			}
			__CLEAR_RESULT

			%%
			;
		};
	};// method

// SQL FUNCTIONS

};

//END OF CLASS CLIPBOARD
