// TYPES FOR CO

#include "types.t"
#include "defines.t"

// This code assumes that spreadsheet_cell::setAttribute accepts spreadsheet id as par_id

// Maximum depth of dependency in expression:
#define EXPR_MAX_DEP 20

// CLASS expression

class expression
{
  inherits 0;

  include_c "http_protocol.h";

#include "../spreadsheet_cell/spreadsheet_acl.t"

  uses "js_code";
  uses "spreadsheet_cell";
  uses "spreadsheet_column";
  uses "expression_cell";
  uses "expression_range";
  uses "clipboard";

  container_class "spreadsheet";
  container_column_name "id_spreadsheet";

 // PROPERTIES
   area sys;

index scrid
{
 unique 1;
 column id_spreadsheet;
 column id_column;
 column id_row;
};

prop version
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

prop deleted
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

prop id_spreadsheet
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

prop id_column
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

prop id_row
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

prop expression_parsed
{
  type CO_STRING;
  default "''";
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

// SQL FUNCTIONS
   function_sql fillV
   {
     doc "copy parsed expressions (gor given cell) down to given row";
     params '(int,int,int,int)';
     returns float4;
   S
   %%
   @"
DECLARE
    _id_spreadsheet ALIAS FOR \$1;
    _id_column ALIAS FOR \$2;
    _id_row ALIAS FOR \$3;
    _id_row_last ALIAS FOR \$4;
    _nr_row_first INTEGER;
    _nr_row_last INTEGER;
BEGIN
    SELECT row_number INTO _nr_row_first FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row
    ;
    SELECT row_number INTO _nr_row_last FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row_last
    ;
    DELETE FROM co_sys_expression
        WHERE EXISTS(
            SELECT * FROM co_sys_spreadsheet_row  AS ROW
                WHERE
                ROW.id = id_row AND
                ROW.row_number > _nr_row_first AND
                ROW.row_number <= _nr_row_last
            )
	    AND id_column = _id_column
    ;
    /* referred column id never change; ref. row id may change */
    INSERT INTO co_sys_expression
        SELECT
        RANGE.version,
        RANGE.deleted,
        RANGE.id_spreadsheet,
        RANGE.id_column,
        ROW.id,
        RANGE.expression_parsed

        FROM
            co_sys_expression AS RANGE,
            co_sys_spreadsheet_row AS ROW
        WHERE
            RANGE.id_spreadsheet = _id_spreadsheet AND
            RANGE.id_column = _id_column AND
            RANGE.id_row = _id_row AND
            ROW.spreadsheet_id = _id_spreadsheet AND
            ROW.row_number > _nr_row_first AND
            ROW.row_number <= _nr_row_last
    ;
RETURN 0;

END;
    "
   %%
   ;
   };

 // METHODS

#include "expression_acl.t"

  method set
  {
    id_permission __PERMISSION_UPDATE;
#if 0
    server_event onafter
    {
     execute_jscript
     %%
     ap_write("alert('=>expression::set');");
     %%
     ;
    };
#endif

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param expression_script { type CO_STRING; };
 
    declare language_c
    {
     C
     %%
     char *expression_result;
     const char *context;
     const char *source;
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
    id_spreadsheet = par_id;

    context = ap_psprintf(p, "
var context = new Object();\n
context.column_id = %d;\n
context.row_id = %d;\n
context.spreadsheet_id = %d;\n
",
    id_column, id_row, id_spreadsheet );

    source = ap_psprintf(p, "return parseExpression( '%s' )",
		    cou_buildEncodedJS(cnx, expression_script) );

    @co_call('"sys"','"js_code"', "evaluate", \
	    par_area=>'"sys"', \
	    par_class=>'"js_code"', \
	    par_id=>-2, \
	    prolog=>context, \
	    direct_code=>source, \
	    direct_libs=>'"expression_parser,expression_runtime"', \
	    asp=>0 );

    expression_result = (char*)cou_getReturnValue( cnx, CO_STRING, &err );

    /* set cell value */
    @co_call('"sys"','"spreadsheet_cell"', "setAttribute", \
            par_id => id_spreadsheet, par_area => '"sys"', par_class => '"spreadsheet_cell"', \
            column_id => id_column, row_id => id_row, attribute_name => '"value"', attribute_value => expression_result);


    if (cnx->putJS) {
        snprintf(JSOutBuf, JSBUFSIZE, "w_c_by_id(%d,%d);\n", id_column, id_row);
    	cou_putJS(cnx,JSOutBuf);
    }
     %%
     ;
    };
  };

  method purge
  {
    id_permission __PERMISSION_PURGE;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
     
          __CREATE_QUERY( "
DELETE FROM co_sys_expression
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d ;
DELETE FROM co_sys_expression_cell
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d ;
DELETE FROM co_sys_expression_range
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d ;
",
	  id_spreadsheet,id_column,id_row,
	  id_spreadsheet,id_column,id_row,
	  id_spreadsheet,id_column,id_row)

          __SQL_EXEC

          {
	    const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
	    int object_ids[] = {id_spreadsheet,id_column,id_row};
				  
    	    cou_queueActionX( cnx, "expression", class_names, object_ids,
		CO_OP_DELETE, NULL );
	  }
	    
     %%
     ;
    };
  };

  method purgeExpressionsInRow
  {
    id_permission __PERMISSION_PURGE;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_row { type CO_INTEGER; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    }; 

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          __CREATE_QUERY( "
DELETE FROM co_sys_expression
WHERE id_spreadsheet=%d AND id_row=%d ;
DELETE FROM co_sys_expression_cell
WHERE id_spreadsheet=%d AND id_row=%d ;
DELETE FROM co_sys_expression_range
WHERE id_spreadsheet=%d AND id_row=%d ;
",
	  id_spreadsheet,id_row,
	  id_spreadsheet,id_row,
	  id_spreadsheet,id_row)

          __SQL_EXEC

          {
	    const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
	    int object_ids[] = {id_spreadsheet,0,id_row};
				  
    	    cou_queueActionX( cnx, "expression", class_names, object_ids,
		CO_OP_DELETE, NULL );
	  }
     %%
     ;
    };
  };

  method purgeExpressionsInColumn
  {
    id_permission __PERMISSION_PURGE;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          __CREATE_QUERY( "
DELETE FROM co_sys_expression
WHERE id_spreadsheet=%d AND id_column=%d ;
DELETE FROM co_sys_expression_cell
WHERE id_spreadsheet=%d AND id_column=%d ;
DELETE FROM co_sys_expression_range
WHERE id_spreadsheet=%d AND id_column=%d ;
",
	  id_spreadsheet,id_column,
	  id_spreadsheet,id_column,
	  id_spreadsheet,id_column)

          __SQL_EXEC

          {
	    const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
	    int object_ids[] = {id_spreadsheet,id_column,0};
				  
    	    cou_queueActionX( cnx, "expression", class_names, object_ids,
		CO_OP_DELETE, NULL );
	  }
     %%
     ;
    };
  };

  method setParsed
  {
    id_permission __PERMISSION_UPDATE;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param expression_parsed { type CO_STRING; };

    declare language_c
    {
     C
     %%
      const char *sql_expression_parsed;
      int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
     
	  sql_expression_parsed = cou_buildEncodedSQL(cnx, expression_parsed);

          __CREATE_QUERY( "
UPDATE co_sys_expression
 SET expression_parsed='%s'
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d
",	  sql_expression_parsed, id_spreadsheet, id_column, id_row)

          __SQL_EXEC

	  if (!ok) { /* new expression */
		snprintf(SQLString, SQLBUFSIZE, "
INSERT INTO co_sys_expression
 (id_spreadsheet,id_column,id_row,expression_parsed)
VALUES(%d,%d,%d,'%s')
",		id_spreadsheet, id_column, id_row,sql_expression_parsed);
 

    	        __SQL_EXEC
	  }
	  {
	     const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
	     int object_ids[] = {id_spreadsheet, id_column, id_row };
             cou_queueActionX( cnx, "expression", class_names, object_ids, CO_OP_UPDATE, "expression_parsed", expression_parsed, NULL );
	  }

     %%
     ;
    };
  };

  method eval
  {
    id_permission __PERMISSION_EVALUATE;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param expression_parsed { type CO_STRING; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
     %%
     ;
    };
  };


  method getCellRefs
  {
    id_permission __PERMISSION_SELECT;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER

    declare language_c
    {
     C
     %%
      int nr;
      int ntuples;
      int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
cou_logError( cnx, CO_LOG_DEBUG, "getCellRefs called");

    /* cells referenced */
        __CREATE_QUERY( "
SELECT EXPR.id_column, EXPR.id_row, EXPR.index_ref,
     EXPR.id_ref_column, EXPR.id_ref_row, EXPR.is_ref_column_hard, EXPR.is_ref_row_hard,
     EXPR.is_ref_column_named, EXPR.is_ref_row_named 
FROM co_sys_spreadsheet_row AS TR,co_sys_spreadsheet_column AS TC, co_sys_expression_cell AS EXPR 
WHERE
    TR.spreadsheet_id = %d AND
    TC.spreadsheet_id = %d AND
    EXPR.id_column=TC.id AND
    EXPR.id_row=TR.id ;
", id_spreadsheet, id_spreadsheet)

    __SQL_GET_RESULT

    ntuples = cor_GetTupleCount(cnx, corec);
    for (nr=0; nr<ntuples; nr++) {
	ap_rprintf( r, "arrExpRefCells['%s_%s_%s']='%s,%s,%s,%s,%s,%s';\n",
	    cor_GetValue(cnx, corec,nr,0),cor_GetValue(cnx, corec,nr,1),cor_GetValue(cnx, corec,nr,2),
	    cor_GetValue(cnx, corec,nr,3),cor_GetValue(cnx, corec,nr,4),
	    cor_GetValue(cnx, corec,nr,5),cor_GetValue(cnx, corec,nr,6),cor_GetValue(cnx, corec,nr,7),cor_GetValue(cnx, corec,nr,8)
	);
    }

    __CLEAR_RESULT

    /* ranges referenced */
        __CREATE_QUERY( "
SELECT EXPR.id_column, EXPR.id_row, EXPR.index_ref,
    EXPR.id_ref_column, EXPR.id_ref_row, 
    EXPR.is_ref_column_hard, EXPR.is_ref_row_hard,
    EXPR.is_ref_column_named, EXPR.is_ref_row_named,
    EXPR.id_ref_column2, EXPR.id_ref_row2, 
    EXPR.is_ref_column2_hard, EXPR.is_ref_row2_hard,
    EXPR.is_ref_column2_named, EXPR.is_ref_row2_named
FROM co_sys_spreadsheet_row AS TR,co_sys_spreadsheet_column AS TC, co_sys_expression_range AS EXPR 
WHERE
    TR.spreadsheet_id = %d AND
    TC.spreadsheet_id = %d AND
    EXPR.id_column=TC.id AND
    EXPR.id_row=TR.id ;
", id_spreadsheet, id_spreadsheet)

    __SQL_GET_RESULT

    ntuples = cor_GetTupleCount(cnx, corec);
    for (nr=0; nr<ntuples; nr++) {
	ap_rprintf( r, "arrExpRefRanges['%s_%s_%s']='%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';\n",
	    cor_GetValue(cnx, corec,nr,0),cor_GetValue(cnx, corec,nr,1),cor_GetValue(cnx, corec,nr,2),
	    cor_GetValue(cnx, corec,nr,3),cor_GetValue(cnx, corec,nr,4),cor_GetValue(cnx, corec,nr,5),cor_GetValue(cnx, corec,nr,6),
	    cor_GetValue(cnx, corec,nr,7),cor_GetValue(cnx, corec,nr,8),cor_GetValue(cnx, corec,nr,9),cor_GetValue(cnx, corec,nr,10),
	    cor_GetValue(cnx, corec,nr,11),cor_GetValue(cnx, corec,nr,12),cor_GetValue(cnx, corec,nr,13),cor_GetValue(cnx, corec,nr,14)
	);
    }

    __CLEAR_RESULT

cou_logError( cnx, CO_LOG_DEBUG, "getCellRefs done");
     %%
     ;
    };
  };

  method copyExpressionToClipboard
  {
    id_permission __PERMISSION_GET;
    doc "Put 'movable' string representation of expression into the user's clipboard.";

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
      co_typein_insertClipboardObject *in_params;
      int user_id;
      int ntuples;
      int num_row=0, num_column=0;
      int i;
      const char *serialization;
      int id_spreadsheet;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      id_spreadsheet = par_id;
//TODO: move to expression_cell/range
      user_id = cnx->id_user;
      serialization = ap_psprintf(p, "%d,%d,%d@", id_spreadsheet, id_column, id_row) ;

        __SQL_GET_INTEGER(num_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_row )
	__ASSERT( (num_row>0) , "row with id=%d not found", id_row)
        __SQL_GET_INTEGER(num_column,
            "SELECT column_number FROM co_sys_spreadsheet_column WHERE id=%d",
            id_column )
	__ASSERT( (num_column>0) , "column with id=%d not found", id_column)

      __CREATE_QUERY("
        SELECT 
index_ref, id_ref_spreadsheet, id_ref_row, id_ref_column, 
is_ref_row_hard, is_ref_column_hard,
is_ref_row_named, is_ref_column_named,
REFCOL.column_number-%d,
REFROW.row_number-%d
 FROM co_sys_expression_cell,
 co_sys_spreadsheet_column AS REFCOL,
 co_sys_spreadsheet_row AS REFROW
        WHERE id_column=%d AND id_row=%d AND
REFCOL.id = id_ref_column AND
REFROW.id = id_ref_row
",	num_column, num_row, id_column, id_row)
      __SQL_GET_RESULT

      ntuples = cor_GetTupleCount(cnx, corec);
      if (ntuples>0) for (i=0; i<ntuples; i++) {
	
	if (i==0)
	serialization = ap_psprintf(p, "%s%s,%s,%s,%s,%s,%s,%s,%s,%s,%s", serialization,
	cor_GetValue(cnx, corec,0,0),cor_GetValue(cnx, corec,0,1),cor_GetValue(cnx, corec,0,2),cor_GetValue(cnx, corec,0,3),
	cor_GetValue(cnx, corec,0,4),cor_GetValue(cnx, corec,0,5),cor_GetValue(cnx, corec,0,6),cor_GetValue(cnx, corec,0,7),
	cor_GetValue(cnx, corec,0,8), cor_GetValue(cnx, corec,0,9) ); 
	else
	serialization = ap_psprintf(p, "%s|%s,%s,%s,%s,%s,%s,%s,%s,%s,%s", serialization,
	cor_GetValue(cnx, corec,i,0),cor_GetValue(cnx, corec,i,1),cor_GetValue(cnx, corec,i,2),cor_GetValue(cnx, corec,i,3),
	cor_GetValue(cnx, corec,i,4),cor_GetValue(cnx, corec,i,5),cor_GetValue(cnx, corec,i,6),cor_GetValue(cnx, corec,i,7), 
	cor_GetValue(cnx, corec,i,8), cor_GetValue(cnx, corec,i,9) ); 
      }
      __CLEAR_RESULT

      __CREATE_QUERY("
        SELECT 
index_ref, function,
id_ref_spreadsheet, 
id_ref_row, id_ref_column, 
is_ref_row_hard, is_ref_column_hard,
is_ref_row_named, is_ref_column_named,
REFCOL.column_number-%d,
REFROW.row_number-%d,
id_ref_row2, id_ref_column2, 
is_ref_row2_hard, is_ref_column2_hard,
is_ref_row2_named, is_ref_column2_named,
REFCOL2.column_number-%d,
REFROW2.row_number-%d
 FROM co_sys_expression_range,
 co_sys_spreadsheet_column AS REFCOL,
 co_sys_spreadsheet_row AS REFROW,
 co_sys_spreadsheet_column AS REFCOL2,
 co_sys_spreadsheet_row AS REFROW2
        WHERE id_column=%d AND id_row=%d AND
REFCOL.id = id_ref_column AND
REFROW.id = id_ref_row AND
REFCOL2.id = id_ref_column2 AND
REFROW2.id = id_ref_row2

",	num_column, num_row, num_column, num_row, id_column, id_row)
      __SQL_GET_RESULT

      ntuples = cor_GetTupleCount(cnx, corec);
      if (ntuples>0) for (i=0; i<ntuples; i++) {
	if (i==0)
	serialization = ap_psprintf(p, "%s#%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s", serialization,
	cor_GetValue(cnx, corec,0,0),cor_GetValue(cnx, corec,0,1),cor_GetValue(cnx, corec,0,2),cor_GetValue(cnx, corec,0,3),
	cor_GetValue(cnx, corec,0,4),cor_GetValue(cnx, corec,0,5),cor_GetValue(cnx, corec,0,6),cor_GetValue(cnx, corec,0,7),
	cor_GetValue(cnx, corec,0,8),cor_GetValue(cnx, corec,0,9),cor_GetValue(cnx, corec,0,10),cor_GetValue(cnx, corec,0,11),
	cor_GetValue(cnx, corec,0,12),cor_GetValue(cnx, corec,0,13),cor_GetValue(cnx, corec,0,14),
	cor_GetValue(cnx, corec,0,15),cor_GetValue(cnx, corec,0,16),cor_GetValue(cnx, corec,0,17), cor_GetValue(cnx, corec,0,18) ); 
	else
	serialization = ap_psprintf(p, "%s|%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s", serialization,
	cor_GetValue(cnx, corec,i,0),cor_GetValue(cnx, corec,i,1),cor_GetValue(cnx, corec,i,2),cor_GetValue(cnx, corec,i,3),
	cor_GetValue(cnx, corec,i,4),cor_GetValue(cnx, corec,i,5),cor_GetValue(cnx, corec,i,6),cor_GetValue(cnx, corec,i,7),
	cor_GetValue(cnx, corec,i,8),cor_GetValue(cnx, corec,i,9),cor_GetValue(cnx, corec,i,10),cor_GetValue(cnx, corec,i,11),
	cor_GetValue(cnx, corec,i,12),cor_GetValue(cnx, corec,i,13),cor_GetValue(cnx, corec,i,14),
	cor_GetValue(cnx, corec,i,15),cor_GetValue(cnx, corec,i,16),cor_GetValue(cnx, corec,i,17), cor_GetValue(cnx, corec,i,18) ); 
      }
      else /* no ranges */
	serialization = ap_psprintf(p, "%s#", serialization);

      __CLEAR_RESULT

      __CREATE_QUERY("
        SELECT expression_parsed from co_sys_expression
        WHERE id_column=%d AND id_row=%d", id_column, id_row)
      __SQL_GET_RESULT

      serialization = ap_psprintf(p, "%s$%s", serialization, cor_GetValue(cnx, corec,0,0) ); 

      __CLEAR_RESULT

      @co_call('"sys"','"clipboard"',"insertClipboardObject", \
              serialization => serialization, \
              par_id=>user_id, \
              par_area=>'"sys"', \
              par_class=>'"clipboard"', \
              owner_area=>'"sys"', \
              owner_class=>'"expression"' \
              );

      %%
      ;
    };
  };

 local_method loc_getIdByNumber
 {
    C
    %%
    int loc_getIdByNumber(co_context *cnx, const char* class_name, int id_parent, int number)
    {
	int *int_ptr;
	int err=0;
	pool *p = cnx->request->pool;
	{
        __PARAM_STRUCT_CREATE_getIdByNumber(in_params)
        __PARAM_STRUCT_SET("sys",            in_params, par_area)
        __PARAM_STRUCT_SET(class_name,       in_params, par_class)
        __PARAM_STRUCT_SET(-1,               in_params, par_id)
        __PARAM_STRUCT_SET(id_parent,        in_params, id_parent)
        __PARAM_STRUCT_SET(number,           in_params, number)
        cou_broker( "sys", class_name, "getIdByNumber", cnx, in_params );
	}
	int_ptr = cou_getReturnValue(cnx, CO_INTEGER, &err);
	if (err) return 0;
	if (int_ptr) return *int_ptr;
	return 0; /* got NULL but no error? */
    }
    %%
    ;
  };

 local_method loc_setParsedReference
 {
    C
    %%
    void loc_setParsedReference(co_context *cnx, 
	    int id_spreadsheet, int id_column, int id_row, /* where */
	    int num_column, int num_row, const char *str_cell_ref) {
	int index_ref, id_ref_spreadsheet, id_ref_row, id_ref_column;
	char is_ref_row_hard[1], is_ref_column_hard[1];
	char is_ref_row_named[1], is_ref_column_named[1];
	int rel_column_number, rel_row_number;
	pool *p = cnx->request->pool;

	sscanf(str_cell_ref, "%d,%d,%d,%d,%c,%c,%c,%c,%d,%d",
	    &index_ref, &id_ref_spreadsheet, &id_ref_row, &id_ref_column, 
	    is_ref_row_hard, is_ref_column_hard,
	    is_ref_row_named, is_ref_column_named,
	    &rel_column_number, &rel_row_number );

	if (is_ref_row_hard[0]=='f')
	{
	    id_ref_row = loc_getIdByNumber(cnx, 
		"spreadsheet_row", id_ref_spreadsheet, num_row+rel_row_number);
	}
	if (is_ref_column_hard[0]=='f')
	{
	    id_ref_column = loc_getIdByNumber(cnx, 
		"spreadsheet_column", id_ref_spreadsheet, num_column+rel_column_number);
	}
	{
        __PARAM_STRUCT_CREATE_setReferencedCell(in_params)
        __PARAM_STRUCT_SET(id_spreadsheet,     in_params, par_id)
        __PARAM_STRUCT_SET(id_column,          in_params, id_column)
        __PARAM_STRUCT_SET(id_row,             in_params, id_row)
        __PARAM_STRUCT_SET(index_ref,          in_params, index)
        __PARAM_STRUCT_SET(id_ref_spreadsheet, in_params, id_ref_spreadsheet)
        __PARAM_STRUCT_SET(id_ref_column,      in_params, id_ref_column)
        __PARAM_STRUCT_SET(id_ref_row,         in_params, id_ref_row)
        __PARAM_STRUCT_SET(is_ref_column_hard[0]=='t',  in_params, ref_column_hard)
        __PARAM_STRUCT_SET(is_ref_row_hard[0]=='t',     in_params, ref_row_hard)
        __PARAM_STRUCT_SET(is_ref_column_named[0]=='t', in_params, ref_column_named)
        __PARAM_STRUCT_SET(is_ref_row_named[0]=='t',    in_params, ref_row_named)

        cou_broker( "sys", "expression_cell", "setReferencedCell", cnx, in_params );
	}
    }
    %%
    ;
  };

 local_method loc_setParsedRange
 {
    C
    %%
    void loc_setParsedRange(co_context *cnx,
	    int id_spreadsheet, int id_column, int id_row, /* where */
	    int num_column, int num_row, const char *str_range_ref) {
	int index_ref, id_ref_spreadsheet;
	int id_ref_row, id_ref_column;
	int id_ref_row2, id_ref_column2;
	int function;
	char is_ref_row_hard[2], is_ref_column_hard[2];
	char is_ref_row_named[2], is_ref_column_named[2];
	int rel_column_number, rel_row_number;
 	int rel_column2_number, rel_row2_number;
	pool *p = cnx->request->pool;

	sscanf(str_range_ref, "%d,%d,%d,%d,%d,%c,%c,%c,%c,%d,%d,%d,%d,%c,%c,%c,%c,%d,%d",
	    &index_ref, &function, &id_ref_spreadsheet,
	    &id_ref_row, &id_ref_column, 
	    is_ref_row_hard, is_ref_column_hard,
	    is_ref_row_named, is_ref_column_named,
	    &rel_column_number, &rel_row_number,
	    &id_ref_row2, &id_ref_column2, 
	    is_ref_row_hard+1, is_ref_column_hard+1,
	    is_ref_row_named+1, is_ref_column_named+1,
	    &rel_column2_number, &rel_row2_number );

	if (is_ref_row_hard[0]=='f')
	{
	    id_ref_row = loc_getIdByNumber(cnx, 
		"spreadsheet_row", id_ref_spreadsheet, num_row+rel_row_number);
	}
	if (is_ref_column_hard[0]=='f')
	{
	    id_ref_column = loc_getIdByNumber(cnx, 
		"spreadsheet_column", id_ref_spreadsheet, num_column+rel_column_number);
	}
	if (is_ref_row_hard[1]=='f')
	{
	    id_ref_row2 = loc_getIdByNumber(cnx, 
		"spreadsheet_row", id_ref_spreadsheet, num_row+rel_row2_number);
	}
	if (is_ref_column_hard[1]=='f')
	{
	    id_ref_column2 = loc_getIdByNumber(cnx, 
		"spreadsheet_column", id_ref_spreadsheet, num_column+rel_column2_number);
	}
	{
        __PARAM_STRUCT_CREATE_setReferencedRange(in_params)
        __PARAM_STRUCT_SET(id_spreadsheet,     in_params, par_id)
        __PARAM_STRUCT_SET(id_column,          in_params, id_column)
        __PARAM_STRUCT_SET(id_row,             in_params, id_row)
        __PARAM_STRUCT_SET(index_ref,          in_params, index)
        __PARAM_STRUCT_SET(function,           in_params, function)
        __PARAM_STRUCT_SET(id_ref_spreadsheet, in_params, id_ref_spreadsheet)
        __PARAM_STRUCT_SET(id_ref_column,      in_params, id_ref_column)
        __PARAM_STRUCT_SET(id_ref_row,         in_params, id_ref_row)
        __PARAM_STRUCT_SET(is_ref_column_hard[0]=='t',  in_params, ref_column_hard)
        __PARAM_STRUCT_SET(is_ref_row_hard[0]=='t',     in_params, ref_row_hard )
        __PARAM_STRUCT_SET(is_ref_column_named[0]=='t', in_params, ref_column_named)
        __PARAM_STRUCT_SET(is_ref_row_named[0]=='t',    in_params, ref_row_named )
        __PARAM_STRUCT_SET(id_ref_column2,              in_params, id_ref_column2)
        __PARAM_STRUCT_SET(id_ref_row2,                 in_params, id_ref_row2)
        __PARAM_STRUCT_SET(is_ref_column_hard[1]=='t',  in_params, ref_column2_hard)
        __PARAM_STRUCT_SET(is_ref_row_hard[1]=='t',     in_params, ref_row2_hard)
        __PARAM_STRUCT_SET(is_ref_column_named[1]=='t', in_params, ref_column2_named)
        __PARAM_STRUCT_SET(is_ref_row_named[1]=='t',    in_params, ref_row2_named)

        cou_broker( "sys", "expression_range", "setReferencedRange", cnx, in_params );
	}
    }
    %%
    ;
  };

  method setExpressionFromClipboard
  {
    id_permission __PERMISSION_UPDATE;
    doc "Creates or updates expression from user's clipboard.";
    doc "New copy of the expression previously saved with copyExpressionToClipboard is placed in new 'virtual' cell.";

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
      int user_id;
      const char *serialization;
      int num_row=0, num_column=0;
      int id_spreadsheet;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      id_spreadsheet = par_id;
      user_id = cnx->id_user;

        __SQL_GET_INTEGER(num_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_row )
        __ASSERT( (num_row>0) , "column with id=%d not found", id_row)
        __SQL_GET_INTEGER(num_column,
            "SELECT column_number FROM co_sys_spreadsheet_column WHERE id=%d",
            id_column )
        __ASSERT( (num_column>0) , "column with id=%d not found", id_column)

      @co_call('"sys"','"clipboard"',"retrieveClipboardObject", \
        par_area=>'"sys"', par_class=>'"clipboard"', par_id=>user_id, \
        owner_area=>'"sys"', owner_class=>'"expression"' );

      serialization =  (const char *) cou_getReturnValue(cnx, CO_STRING, &err);

    /* parsing */
    while ((*serialization) && (*serialization++ != '@')); /* skip original cell position */
    {
	char *s_ptr = ap_pstrdup(p, serialization);
	char *s_ptr2;
	char *s_cell_ref, *s_range_ref, *expression_parsed;

	s_cell_ref = s_ptr;
	/* go to range refs and add cell refs  */
	while (*s_ptr) { 
	    if (*s_ptr=='#') { 
		*s_ptr='\0';
		if (*s_cell_ref) loc_setParsedReference(cnx,
		    id_spreadsheet, id_column, id_row,
		    num_column, num_row, s_cell_ref);
		s_ptr++;
		break;
	    } 
	    else 
	    if (*s_ptr=='|') {
		*s_ptr='\0';
		s_ptr++;
		if (*s_cell_ref) loc_setParsedReference(cnx, 
		    id_spreadsheet, id_column, id_row,
		    num_column, num_row, s_cell_ref);
		s_cell_ref = s_ptr;
	    } 
	    else s_ptr++;
	};
	s_range_ref = s_ptr;

	/* go to expression text and add range refs */
	while (*s_ptr) { 
	    if (*s_ptr=='$') { 
		*s_ptr='\0';
		if (*s_range_ref) loc_setParsedRange(cnx, 
		    id_spreadsheet, id_column, id_row,
		    num_column, num_row, s_range_ref);
		s_ptr++;
		break;
	    } 
	    else 
	    if (*s_ptr=='|') {
		*s_ptr='\0';
		s_ptr++;
		if (*s_range_ref) loc_setParsedRange(cnx, 
		    id_spreadsheet, id_column, id_row,
		    num_column, num_row, s_range_ref);
		s_range_ref = s_ptr;
	    } 
	    else s_ptr++;
	};
	expression_parsed = s_ptr;

        @co_call('"sys"','"expression"',"setParsed", \
            par_id=>id_spreadsheet, \
            id_row=>id_row, id_column=>id_column, \
	    expression_parsed=>expression_parsed );

        @co_call('"sys"','"expression"',"evalRectangle", \
	    par_id=>id_spreadsheet, \
	    id_column=>id_column,id_row=>id_row, \
	    id_last_column=>id_column,id_last_row=>id_row);

    }

      %%
      ;
    };

  };

  method evalDependent
  {
    id_permission __PERMISSION_EVALUATE;
    doc "Finds and evaluates expressions depending on given cell.";
//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
        __CREATE_QUERY( "
DELETE FROM co_sys_expression_deps WHERE id_session = %d ;

INSERT INTO co_sys_expression_deps
SELECT DISTINCT %d as id_session, id_column as c_id, id_row as r_id, 1 as n 
FROM co_sys_expression_cell AS CELL
WHERE CELL.id_ref_column = %d AND
  CELL.id_ref_row = %d
;
INSERT INTO co_sys_expression_deps
SELECT DISTINCT %d as id_session, RANGE.id_column as c_id, RANGE.id_row as r_id,
         1 AS n
  FROM co_sys_spreadsheet_row AS ROWSRC, co_sys_spreadsheet_column AS COLSRC,
    co_sys_expression_range AS RANGE,
    co_sys_spreadsheet_row AS ROWREF1, co_sys_spreadsheet_column AS COLREF1,
    co_sys_spreadsheet_row AS ROWREF2, co_sys_spreadsheet_column AS COLREF2
  WHERE
    ROWSRC.id = %d AND
    COLSRC.id = %d AND
    ROWREF1.id=RANGE.id_ref_row AND
    COLREF1.id=RANGE.id_ref_column AND
    ROWREF2.id=RANGE.id_ref_row2 AND
    COLREF2.id=RANGE.id_ref_column2 AND
    ROWSRC.spreadsheet_id = ROWREF1.spreadsheet_id AND -- range within the same spreadsheet as our cell
    ROWSRC.row_number >= ROWREF1.row_number AND
    ROWSRC.row_number <= ROWREF2.row_number AND
    COLSRC.column_number >= COLREF1.column_number AND
    COLSRC.column_number <= COLREF2.column_number
;
",	cnx->id_session, 
	cnx->id_session, id_column, id_row, 
	cnx->id_session, id_row, id_column)

    	__SQL_EXEC

    {
	int n;
	int cell_deps, range_deps;
	int loops; /* loops in expression? */

	for(n=1; n<EXPR_MAX_DEP; n++) {
    	    __CREATE_QUERY( "
  INSERT INTO co_sys_expression_deps 
  SELECT DISTINCT DEP.id_session as id_session, CELL.id_column as c_id, CELL.id_row as r_id, DEP.n+1 as n
  FROM co_sys_expression_deps AS DEP, co_sys_expression_cell AS CELL
  WHERE
    DEP.id_session = %d AND
    DEP.n = %d  AND -- used in loop
    CELL.id_ref_column = DEP.c_id AND
    CELL.id_ref_row = DEP.r_id
 ;
", 
	    cnx->id_session, n)

    	    __SQL_EXEC
	    cell_deps = ok; /* atoi(PQcmdTuple(res)); */

    	    __CREATE_QUERY( "
  INSERT INTO co_sys_expression_deps 
  SELECT DISTINCT DEP.id_session as id_session, RANGE.id_column as c_id, RANGE.id_row as r_id, DEP.n+1 as n
  FROM co_sys_expression_deps AS DEP, co_sys_spreadsheet_row AS ROWSRC, co_sys_spreadsheet_column AS COLSRC,
    co_sys_expression_range AS RANGE,
    co_sys_spreadsheet_row AS ROWREF1, co_sys_spreadsheet_column AS COLREF1,
    co_sys_spreadsheet_row AS ROWREF2, co_sys_spreadsheet_column AS COLREF2
  WHERE
    DEP.id_session = %d AND
    ROWSRC.id=DEP.r_id AND
    COLSRC.id=DEP.c_id AND
    ROWREF1.id=RANGE.id_ref_row AND
    COLREF1.id=RANGE.id_ref_column AND
    ROWREF2.id=RANGE.id_ref_row2 AND
    COLREF2.id=RANGE.id_ref_column2 AND
    ROWSRC.spreadsheet_id = ROWREF1.spreadsheet_id AND -- range within the same spreadsheet as our cell
    ROWSRC.row_number >= ROWREF1.row_number AND
    ROWSRC.row_number <= ROWREF2.row_number AND
    COLSRC.column_number >= COLREF1.column_number AND
    COLSRC.column_number <= COLREF2.column_number AND
    DEP.n = %d -- used in loop
  ;
", 
	    cnx->id_session, n)
    	    __SQL_EXEC
	    range_deps = ok; /* atoi(PQcmdTuple(res)); */

	    if( (cell_deps==0) && (range_deps==0) ) break;

	    if((n+1)>=EXPR_MAX_DEP) {
		/* last loop in for... and we are still getting dependent cells? */
		cou_error( cnx, "Max dependency depth exceeded");
	    }

	    loops = 0;
    	    __SQL_GET_INTEGER( loops, "
SELECT count(*) FROM co_sys_expression_deps AS DEP
WHERE DEP.id_session=%d AND
   DEP.c_id=%d AND DEP.r_id=%d
;
", 
	    cnx->id_session, id_column, id_row)

	    if (loops>0) {
		cou_error( cnx, "Expression Loop occured");
	    }
	}
    } /* deps generation */

    __CREATE_QUERY( "
SELECT DEP.c_id, DEP.r_id, EXPR.expression_parsed, EXPR.id_spreadsheet
FROM co_sys_expression_deps AS DEP, co_sys_expression AS EXPR
WHERE DEP.id_session=%d AND
   EXPR.id_column = DEP.c_id AND EXPR.id_row=DEP.r_id
GROUP BY DEP.c_id, DEP.r_id, EXPR.expression_parsed, EXPR.id_spreadsheet
ORDER BY MAX(n)
;
", 
    cnx->id_session)

    __SQL_GET_RESULT

cou_logError(cnx, CO_LOG_DEBUG, "ntuples=%d", cor_GetTupleCount(cnx, corec));
{ 
    int i;
    if (cor_GetTupleCount(cnx, corec)>0) for (i=0; i<cor_GetTupleCount(cnx, corec); i++) {
	int dep_id_spreadsheet;
	int dep_id_column;
	int dep_id_row;
	const char *expression_result;
	const char *expression_script;
	const char *source;
        const char *context;

cou_logError(cnx, CO_LOG_DEBUG, "%s,%s,%s,%s",
	cor_GetValue(cnx, corec,i,0),cor_GetValue(cnx, corec,i,1),cor_GetValue(cnx, corec,i,2),cor_GetValue(cnx, corec,i,3));

	dep_id_spreadsheet = atoi(cor_GetValue(cnx, corec,i,3));
	dep_id_column = atoi(cor_GetValue(cnx, corec,i,0));
	dep_id_row = atoi(cor_GetValue(cnx, corec,i,1));
	expression_script = cor_GetValue(cnx, corec,i,2);
	context = ap_psprintf(p, "
var context = new Object();\n
context.column_id = %d;\n
context.row_id = %d;\n
context.spreadsheet_id = %d;
",
	dep_id_column, dep_id_row, dep_id_spreadsheet );

	source = ap_psprintf(p, "return %s", ++expression_script); /* with beginning '=' */

	@co_call('"sys"','"js_code"', "evaluate", \
	    par_area=>'"sys"', \
	    par_class=>'"js_code"', \
	    par_id=>-2, \
	    prolog=>context, \
	    direct_code=>source, \
	    direct_libs=>'"expression_runtime"', \
	    asp=>0 );

	expression_result = (char*)cou_getReturnValue( cnx, CO_STRING, &err );

	/* set cell value */
	@co_call('"sys"', '"spreadsheet_cell"', "setAttribute", \
	          par_id => dep_id_spreadsheet, \
	          par_area => '"sys"', \
	          par_class => '"spreadsheet_cell"', \
	          column_id => dep_id_column, \
	          row_id => dep_id_row, \
	          attribute_name => '"value_noeval"' , \
	          attribute_value => expression_result \
                 );

        if (cnx->putJS) {
	    snprintf(JSOutBuf, JSBUFSIZE, "w_c_by_id(%d,%d);\n", dep_id_column, dep_id_row);
    	    cou_putJS(cnx,JSOutBuf);
        }
    } /* for ... */
}
    __CLEAR_RESULT

     %%
     ;
    };
  };

  method evalRectangle
  {
    id_permission __PERMISSION_EVALUATE;
    doc "Finds and evaluates expressions belonging to and depending on given rectangle.";
    doc "Supported : dependencies nested up to level EXPR_MAX_DEP (set in expression.t)";
    doc "By default, if id of row/column is not specified, rectangle is extended";
    doc "so that by specifying id_spreadsheet only you force evaluation for complete spreadsheet.";
//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; optional 1; default 'id_column=0'; };
    param id_row { type CO_INTEGER; optional 1; default 'id_row=0'; };
    param id_last_column { type CO_INTEGER; optional 1; default 'id_last_column=0'; };
    param id_last_row { type CO_INTEGER; optional 1; default 'id_last_row=0'; };

    declare language_c
    {
      C
      %%
	int num_first_column = 0;
	int num_last_column = 0;
	int num_first_row = 0;
	int num_last_row = 0;
        int id_spreadsheet;
      %%
      ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
	if (id_column>0) {
	    __SQL_GET_INTEGER(num_first_column, 
		"SELECT column_number FROM co_sys_spreadsheet_column WHERE id=%d",
		id_column )
    	    __ASSERT( (num_first_column>0) , "column with id=%d not found", id_column)
	}
	if (id_row>0) {
	    __SQL_GET_INTEGER(num_first_row, 
		"SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
		id_row )
    	    __ASSERT( (num_first_row>0) , "column with id=%d not found", id_row)
	}
	if (id_last_column>0) {
	    __SQL_GET_INTEGER(num_last_column, 
		"SELECT column_number FROM co_sys_spreadsheet_column WHERE id=%d",
	        id_last_column )
    	    __ASSERT( (num_last_column>0) , "column with id=%d not found", id_last_column)
	}
	if (id_last_row>0) {
	    __SQL_GET_INTEGER(num_last_row, 
		"SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
		id_last_row )
    	    __ASSERT( (num_last_row>0) , "row with id=%d not found", id_last_row)
	}
	if (num_last_column==0) {
	    __SQL_GET_INTEGER(num_last_column, 
		"SELECT max(column_number) FROM co_sys_spreadsheet_column WHERE spreadsheet_id=%d",
		id_spreadsheet )
	}
	if (num_last_row==0) {
	    __SQL_GET_INTEGER(num_last_row, 
		"SELECT max(row_number) FROM co_sys_spreadsheet_row WHERE spreadsheet_id=%d",
		id_spreadsheet )
	}
        __CREATE_QUERY( "
DELETE FROM co_sys_expression_deps_r WHERE id_session = %d ;

INSERT INTO co_sys_expression_deps_r
SELECT DISTINCT %d as id_session, COLREF.id as c_id, ROWREF.id as r_id, 
1 as n, COLREF.id as seed_c_id, ROWREF.id as seed_r_id
FROM 
    co_sys_spreadsheet_row AS ROWREF, co_sys_spreadsheet_column AS COLREF
WHERE COLREF.spreadsheet_id=%d AND COLREF.column_number>=%d AND COLREF.column_number<=%d AND
  ROWREF.spreadsheet_id=%d AND ROWREF.row_number>=%d AND ROWREF.row_number<=%d
;
",	cnx->id_session, 
	cnx->id_session, 
	    id_spreadsheet, num_first_column, num_last_column, 
	    id_spreadsheet, num_first_row, num_last_row)

    	__SQL_EXEC

    {
	int n;
	int cell_deps, range_deps;
	int loops; /* loops in expression? */

	for(n=1; n<EXPR_MAX_DEP; n++) {
    	    __CREATE_QUERY( "
  INSERT INTO co_sys_expression_deps_r 
  SELECT DISTINCT DEP.id_session as id_session, CELL.id_column as c_id, CELL.id_row as r_id, 
    DEP.n+1 as n, DEP.seed_c_id as seed_c_id, DEP.seed_r_id as seed_r_id
  FROM co_sys_expression_deps_r AS DEP, co_sys_expression_cell AS CELL
  WHERE
    DEP.id_session = %d AND
    DEP.n = %d  AND -- used in loop
    CELL.id_ref_column = DEP.c_id AND
    CELL.id_ref_row = DEP.r_id
 ;
", 
	    cnx->id_session, n)

    	    __SQL_EXEC
	    cell_deps = ok; /* atoi(PQcmdTuple(res)); */

    	    __CREATE_QUERY( "
  INSERT INTO co_sys_expression_deps_r 
  SELECT DISTINCT DEP.id_session as id_session, RANGE.id_column as c_id, RANGE.id_row as r_id, 
    DEP.n+1 as n, DEP.seed_c_id as seed_c_id, DEP.seed_r_id as seed_r_id
  FROM co_sys_expression_deps_r AS DEP, co_sys_spreadsheet_row AS ROWSRC, co_sys_spreadsheet_column AS COLSRC,
    co_sys_expression_range AS RANGE,
    co_sys_spreadsheet_row AS ROWREF1, co_sys_spreadsheet_column AS COLREF1,
    co_sys_spreadsheet_row AS ROWREF2, co_sys_spreadsheet_column AS COLREF2
  WHERE
    DEP.id_session = %d AND
    ROWSRC.id=DEP.r_id AND
    COLSRC.id=DEP.c_id AND
    ROWREF1.id=RANGE.id_ref_row AND
    COLREF1.id=RANGE.id_ref_column AND
    ROWREF2.id=RANGE.id_ref_row2 AND
    COLREF2.id=RANGE.id_ref_column2 AND
    ROWSRC.spreadsheet_id = ROWREF1.spreadsheet_id AND -- range within the same spreadsheet as our cell
    ROWSRC.row_number >= ROWREF1.row_number AND
    ROWSRC.row_number <= ROWREF2.row_number AND
    COLSRC.column_number >= COLREF1.column_number AND
    COLSRC.column_number <= COLREF2.column_number AND
    DEP.n = %d -- used in loop
  ;
", 
	    cnx->id_session, n)
    	    __SQL_EXEC
	    range_deps = ok; /* atoi(PQcmdTuple(res)); */

	    if( (cell_deps==0) && (range_deps==0) ) break;

	    if((n+1)>=EXPR_MAX_DEP) {
		/* last loop in for... and we are still getting dependent cells? */
		cou_error( cnx, "Max dependency depth exceeded");
	    }

	    /* check for loops: does cell depends on itself? */
	    loops = 0;
    	    __SQL_GET_INTEGER( loops, "
SELECT count(*) FROM co_sys_expression_deps_r AS DEP
WHERE 
    DEP.id_session=%d AND 
    DEP.n>1 AND
    DEP.seed_c_id = DEP.c_id AND DEP.seed_r_id = DEP.r_id
;
", 
	    cnx->id_session)

	    if (loops>0) {
		cou_error( cnx, "Expression Loop occured");
	    }

	} /* loop */
	//if (n>=EXPR_MAX_DEP) cou_error( cnx, "Unsupported dependency depth (nesting level too big)");
    } /* n was visible */

        __CREATE_QUERY( "
SELECT DEP.c_id, DEP.r_id, EXPR.expression_parsed, EXPR.id_spreadsheet
FROM co_sys_expression_deps_r AS DEP, co_sys_expression AS EXPR
WHERE DEP.id_session=%d AND
   EXPR.id_column = DEP.c_id AND EXPR.id_row=DEP.r_id
GROUP BY DEP.c_id, DEP.r_id, EXPR.expression_parsed, EXPR.id_spreadsheet
ORDER BY MAX(n)
;
", 
	cnx->id_session)

    __SQL_GET_RESULT

cou_logError(cnx, CO_LOG_DEBUG, "ntuples=%d", cor_GetTupleCount(cnx, corec));
{ 
    int i;
    if (cor_GetTupleCount(cnx, corec)>0) for (i=0; i<cor_GetTupleCount(cnx, corec); i++) {
	int dep_id_spreadsheet;
	int dep_id_column;
	int dep_id_row;
	const char *expression_result;
	const char *expression_script;
	const char *source;
	const char *context;

cou_logError(cnx, CO_LOG_DEBUG, "%s,%s,%s,%s",
	cor_GetValue(cnx, corec,i,0),cor_GetValue(cnx, corec,i,1),cor_GetValue(cnx, corec,i,2),cor_GetValue(cnx, corec,i,3));

	dep_id_spreadsheet = atoi(cor_GetValue(cnx, corec,i,3));
	dep_id_column = atoi(cor_GetValue(cnx, corec,i,0));
	dep_id_row = atoi(cor_GetValue(cnx, corec,i,1));
	expression_script = cor_GetValue(cnx, corec,i,2);

	context = ap_psprintf(p, "
var context = new Object();\n
context.column_id = %d;\n
context.row_id = %d;\n
context.spreadsheet_id = %d;\n
",
	dep_id_column, dep_id_row, dep_id_spreadsheet );

	source = ap_psprintf(p, "return %s", ++expression_script); /* skips '=' */

	@co_call('"sys"','"js_code"', "evaluate", \
	    par_area=>'"sys"', \
	    par_class=>'"js_code"', \
	    par_id=>-2, \
	    prolog=>context, \
	    direct_code=>source, \
	    direct_libs=>'"expression_runtime"', \
	    asp=>0 );

	expression_result = (char*)cou_getReturnValue( cnx, CO_STRING, &err );

/*
	expression_result = co_priv_callJS(cnx, 
	    dep_id_spreadsheet, dep_id_column, dep_id_row, 
	    "evalExpression", cor_GetValue(cnx, corec,i,2) );
*/
	/* set cell value */
	@co_call('"sys"', '"spreadsheet_cell"', "setAttribute", \
	         par_id => dep_id_spreadsheet, \
	         par_area => '"sys"' , \
	         par_class => '"spreadsheet_cell"' , \
	         column_id => dep_id_column, \
	         row_id => dep_id_row, \
	         attribute_name => '"value_noeval"' , \
	         attribute_value => expression_result \
                );

/*
	cparams.par_id = dep_id_spreadsheet;
	cparams.par_area = "sys";
	cparams.par_class = "spreadsheet_cell";
	cparams.column_id = dep_id_column;
	cparams.row_id = dep_id_row;
	cparams.attribute_name = "value_noeval";
	cparams.attribute_value = expression_result;
	cou_broker( "sys", "spreadsheet_cell", "setAttribute", cnx, &cparams );
*/

        if (cnx->putJS) {
	    snprintf(JSOutBuf, JSBUFSIZE, "w_c_by_id(%d,%d);\n", dep_id_column, dep_id_row);
    	    cou_putJS(cnx,JSOutBuf);
        }
    } /* for ... */
}
    __CLEAR_RESULT

     %%
     ;
    };
  };

  method fillDownExpression
  {
    id_permission 3;

    /* original cell */
//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    /* fill down until row... */
    param id_last_row { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
        int num_first_row=0, num_last_row=0;
        int ntuples, i;
        int id_spreadsheet;
      %%
      ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          __CREATE_QUERY("
SELECT co_sys_expression_fillV(%d,%d,%d,%d);
",      id_spreadsheet, id_column, id_row, id_last_row)

          __SQL_GET_RESULT
          __CLEAR_RESULT

// always copy references with the same arguments
	cou_broker( "sys", "expression_cell", "fillDownCellRef", cnx, struct_in );
	cou_broker( "sys", "expression_range", "fillDownRangeRef", cnx, struct_in );

	@co_call('"sys"','"expression"','evalRectangle',par_id=>id_spreadsheet,id_column=>id_column,id_row=>id_row,id_last_column=>id_column,id_last_row=>id_last_row);

// TODO: send actions
        __SQL_GET_INTEGER(num_first_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_row )
        __ASSERT( (num_first_row>0) , "row with id=%d not found", id_row)
        __SQL_GET_INTEGER(num_last_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_last_row )
        __ASSERT( (num_last_row>0) , "row with id=%d not found", id_last_row)

        __CREATE_QUERY("
SELECT
id_row,
expression_parsed
FROM co_sys_expression AS EXPR, co_sys_spreadsheet_row AS ROW
WHERE EXPR.id_spreadsheet=%d AND EXPR.id_column=%d AND
    EXPR.id_row = ROW.id AND ROW.row_number>%d AND ROW.row_number<=%d
",      id_spreadsheet, id_column,
        num_first_row, num_last_row)

        __SQL_GET_RESULT
        ntuples = cor_GetTupleCount(cnx, corec);
        {
            const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
            int object_ids[] = {id_spreadsheet,id_column,id_row};
            int id_row_current;
            for (i=0; i<ntuples; i++) {
                id_row_current = atoi(cor_GetValue(cnx, corec, i, 0));
                object_ids[2] = id_row_current;
                cou_queueActionX( cnx, "expression", class_names, object_ids,CO_OP_UPDATE, "expression_parsed", cor_GetValue(cnx, corec, i, 1), NULL );
            }
        }
        __CLEAR_RESULT

     %%
     ;
    };
  };

};