// TYPES FOR CO

#include "types.t"
#include "defines.t"

// CLASS file

class file
{
  inherits generic_acl;

  container_class "";
  container_column_name "";

  area sys;
  area sec;
  area usr;

#include "../../include/check_acl.t"

  include_c "http_protocol.h";
  include_c "co_mod.h";
  include_c "../coffice/mod_coffice.h";

  uses "generic";

/* mapping files into memory */
  include_c "stdlib.h";
  include_c "unistd.h";
  include_c "sys/mman.h";
  include_c "fcntl.h";

 // PROPERTIES
#include "generic_acl_props.t"

 prop file_version
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 0;
   foreign '';
   foreign_prop '';
   position 100;
  };


 // METHODS

  method getFile
  {
    doc "Sends file content to the client, including header with Content Type.";
    doc "Does not return file: use getFileAsString in scripts and/or methods.";
    id_permission __PERMISSION_GET;

    __PARAMS_HEADER

    declare language_c
    {
     C
     %%
	const char *file_path, *partial_path;
	FILE *fh;
     %%
     ;
    };

    body language_c
    {
     C
     %%
//
	@co_call(par_area,par_class,'getPath',par_id=>par_id,par_area=>par_area,par_class=>par_class);
	partial_path = cou_getReturnValue( cnx, CO_STRING, &err );
	file_path = ap_psprintf( p, "%s/%s", cou_getCODataPath() , partial_path );

	cou_logError( cnx, CO_LOG_DEBUG, "filename %s",file_path);


	@co_call(par_area,par_class,'getContentType',par_id=>par_id,par_area=>par_area,par_class=>par_class);
	cnx->request->content_type = (char*)cou_getReturnValue( cnx, CO_STRING, &err );

	ap_send_http_header(r);

	fh = ap_pfopen( p, file_path, "r" );

	if (!fh) 
	{
	  __SEND_MESG_DEBUG("filename %s",file_path)
	  cou_error( cnx, "Could not open file");
	}
	
	ap_send_fd( fh, cnx->request ); /* returns total of bytes sent */
	ap_pfclose(p,fh);

	__SEND_MESG_DEBUG("getFile")
     %%
     ;
    };
  };

  method getFileAsString
  {
    doc "Retrieves (0-ended) copy of the file content (as 0-ended string).";
    doc "Does not generate actions nor JavaScript.";
    id_permission __PERMISSION_GET;
    __PARAMS_HEADER

    declare language_c
    {
     C
     %%
	const char *pathname, *partial_path;
	int fd;
	void *map;
	struct stat stat_buf;
	size_t length;
     %%
     ;
    };

    body language_c
    {
     C
     %%
//
	@co_call(par_area,par_class,'getPath',par_id=>par_id,par_area=>par_area,par_class=>par_class);
	partial_path = cou_getReturnValue( cnx, CO_STRING, &err );
	pathname = ap_psprintf( p, "%s/%s", cou_getCODataPath() , partial_path );

        fd = ap_popenf(p, pathname, O_RDONLY, 0);

	if (fd == -1) 
	{
	  __SEND_MESG_DEBUG("Could not open file: %s", pathname)
	  cou_error( cnx, "Could not open file");
	}

	if (fstat(fd, &stat_buf)==-1) /* can't stat file */
	{
	  __SEND_MESG_DEBUG("Could not stat file: %s", pathname)
	  cou_error( cnx, "Could not stat file");
	}
        length = stat_buf.st_size+1;

	map=mmap( 0, length, PROT_READ, MAP_SHARED, fd, 0);

	__SEND_MESG_DEBUG("MMapped %s", pathname)

	cou_setReturnValue(cnx, 
	    ap_pstrndup( p, map, length ),
	    CO_STRING);

	munmap(map, length);
	if (ap_pclosef(p, fd) == -1) cou_logError(cnx, CO_LOG_DEBUG, "couldn't close file %s", pathname);

	__SEND_MESG_DEBUG("getFileAsString")
     %%
     ;
    };
  };

  method writeFileFromString
  {
    doc "Writes content to the file, replaces previous content and increases files version.";
    id_permission __PERMISSION_UPDATE;

    __PARAMS_HEADER

    param text { type CO_STRING; };

    declare language_c
    {
     C
     %%
	const char *pathname, *partial_path;
	int fd;
	size_t len;
	int file_version;
	char * str_file_version;
     %%
     ;
    };

    body language_c
    {
     C
     %%
	/* first, it is a new version - and version may change pathname, updating file_version increases version */
    /* TODO: version control!!!! */
//	@co_call(par_area,par_class,'increaseVersion',par_id=>par_id,par_area=>par_area,par_class=>par_class);
	@co_call(par_area,par_class,'getProp',par_id=>par_id,par_area=>par_area,par_class=>par_class, prop_name=>'"file_version"');
	__GET_RETURNED_STRING(str_file_version)
	file_version = atoi(str_file_version);
    str_file_version = ap_psprintf(p, "%d", file_version+1);
	@co_call(par_area,par_class,'update',par_id=>par_id,par_area=>par_area,par_class=>par_class, version=>-1, par_prop=>'"file_version"', new_value=>str_file_version);

	@co_call(par_area,par_class,'getPath',par_id=>par_id,par_area=>par_area,par_class=>par_class);
	partial_path = cou_getReturnValue( cnx, CO_STRING, &err );
	pathname = ap_psprintf( p, "%s/%s", cou_getCODataPath() , partial_path );

	fd = ap_popenf(p, pathname, O_WRONLY | O_TRUNC | O_CREAT, S_IREAD | S_IWRITE);
	if (fd == -1)
	{
	    __SEND_MESG_DEBUG("couldn't open file %s for writing",pathname)
    	    /* abort transaction - back to the old version */
	    cou_error(cnx, "couldn't open file for writing!");
	}
	len = strlen(text);
	if (write(fd, text, len) != len) {
    	    __SEND_MESG_DEBUG("couldn't write to file %s", pathname)
    	    ap_pclosef(p, fd);
	    /* abort transaction - back to the old version */
	    cou_error(cnx, "couldn't save file!");
	}

	if (ap_pclosef(p, fd) == -1) cou_logError(cnx, CO_LOG_DEBUG, "couldn't close file %s", pathname);
		/* should we continue if we can't close? */

	__SEND_MESG_DEBUG("writeFileFromString")

	// for html compatibility
     if(cnx->client_mode==CO_CLIENT_HTML)
	{
		ap_rprintf(r, "Text changed");
	}

     %%
     ;
    };
  };

  method getPath
  {
    doc "Returns partial path to the file (path relative data root; file name included).";
    doc "You will want to overwrite this method if you use non-standard layout.";
    id_permission __PERMISSION_GET;
    __PARAMS_HEADER

    body language_c
    {
     C
     %%
	__CREATE_QUERY("
	SELECT id, file_version FROM co_%s_%s WHERE id=%d AND deleted=0;
	", par_area, par_class, par_id)
	__SQL_GET_RESULT

	if (ok!=1) cou_error( cnx, "No record found or multiple records");
	/* get: id, version */
	cou_setReturnValue( cnx, 
	    ap_psprintf( p, "%s/%s/%s.%s", par_area, par_class,
		cor_GetValue(cnx, corec, 0, 0), cor_GetValue(cnx, corec, 0, 1)),
	    CO_STRING
	);
	__CLEAR_RESULT
	__SEND_MESG_DEBUG("getFile")

     %%
     ;
    };
  };

  method getContentType
  {
    doc "Returns string containing content type: Application/octet-stream.";
    doc "You will want to overwrite this method.";
    id_permission __PERMISSION_GET;
    __PARAMS_HEADER

    body language_c
    {
     C
     %%
	cou_setReturnValue( cnx, 
	    ap_pstrdup(p,"Application/octet-stream"),
	    CO_STRING
	);

	__SEND_MESG_DEBUG("getContentType")

     %%
     ;
    };
  };

};
