
#include "../../include/types.t"
#include "../../include/defines.t"

class generic_acl
{
//   area "_";
   area "sec";
   area "sys";
   area "usr";

   container_class '';
   container_column_name '';   

#include "../../include/check_acl.t"
    
   inherits generic;

   uses "generic";

#include "../../include/generic_acl_props.t" 
      
   method check_acl
   {
     doc "This is very important method that checks acl. Whole security system depends on this method!";
     doc "You will want to overwrite this method if you have non-standard ACL checking.";
     id_permission 0;
     
     __PARAMS_HEADER
     
     body language_c
     {
       C
       %%
        __SEND_MESG_DEBUG("check_acl in area '%s': class='%s' id_permission=%d par_id=%d \n", par_area, par_class, cnx->id_permission, par_id)

       __SQL_GET_INTEGER(return_int,"SELECT co_%s_%s_check_acl(%d,%d,%d)",
                           par_area, par_class, par_id, cnx->id_permission, cnx->id_role)
			   
        __RETURN_INTEGER(return_int)
       %%
      ;
     };
   };      

   method check_acl_class
   {
     doc "This is very important method that checks acl. Whole security system depends on this method!";
     doc "You will want to overwrite this method if you have non-standard ACL checking.";
     id_permission 0;
     
     __PARAMS_HEADER_CLASS
     
     body language_c
     {
       C
       %%
       __SEND_MESG_DEBUG("check_acl_class in area '%s': class='%s' id_permission=%d id_role=%d ", par_area, par_class, cnx->id_permission, cnx->id_role)  


       __SQL_GET_INTEGER(return_int,"SELECT co_%s_%s_check_acl_class(%d,%d)",
                           par_area, par_class, cnx->id_permission, cnx->id_role)
			   
        __RETURN_INTEGER(return_int)
  
      %%
      ;
     };
   };      


	method testACL
	{
		doc "this is public version of checkACL, that you can use to check the acl value";

		id_permission __PERMISSION_GET_ACL;
		sec_type __METHOD_SEC_TYPE_OBJECT;
    
		__PARAMS_HEADER_WITH_CALLBACK

		param id_permission
		{
			doc "Id of permission that you want to check";
			type CO_INTEGER;
		};
    
		param id_role
		{
			doc "Id of role that you want to check";
			type CO_INTEGER;
		};
    
		declare language_c
		{
			C "int id_permission_old;";
			C "int id_role_old;";
			C "int acl;";
			C "int check_security;";
		};

		body language_c
		{
			C
			%%
			id_permission_old = cnx->id_permission;
			id_role_old = cnx->id_role;
			check_security = cnx->check_security;
			cnx->id_permission = id_permission;     
			cnx->id_role = id_role;
			cnx->check_security = 0;
            
			@co_call(par_area, par_class, "check_acl", par_area=>par_area, par_class=>par_class, par_id=>par_id); 

			cnx->id_permission = id_permission_old;
			cnx->id_role = id_role_old;
			cnx->check_security = check_security;
            
			// we do not have to set return value becouse checkACL does
			// only we have to create JavaScript
			__GET_RETURNED_INTEGER(acl)
			__CREATE_JS_FROM_INTEGER(acl)
			%%
			;
		};
	}; //end of method testACL

	method testClassACL
	{
		doc "this is public version of check_acl_class, that you can use to check the acl value for given class";

		id_permission __PERMISSION_GET_ACL;
		sec_type __METHOD_SEC_TYPE_CLASS;
    
		__PARAMS_HEADER_WITH_CALLBACK

		param id_permission
		{
			doc "Id of permission that you want to check";
			type CO_INTEGER;
		};
    
		param id_role
		{
			doc "Id of role that you want to check";
			type CO_INTEGER;
		};
    
		declare language_c
		{
			C "int id_permission_old;";
			C "int id_role_old;";
			C "int acl;";
			C "int check_security;";
		};

		body language_c
		{
			C
			%%
			id_permission_old = cnx->id_permission;
			id_role_old = cnx->id_role;
			check_security = cnx->check_security;
			cnx->id_permission = id_permission;     
			cnx->id_role = id_role;
			cnx->check_security = 0;
        		__SEND_MESG_DEBUG("testClassACL before call in area '%s': class='%s' id_permission=%d \n", par_area, par_class, cnx->id_permission)
			@co_call(par_area, par_class, "check_acl_class", par_area=>par_area, par_class=>par_class); 

			cnx->id_permission = id_permission_old;
			cnx->id_role = id_role_old;
			cnx->check_security = check_security;
            
			// we do not have to set return value becouse checkACL does
			// only we have to create JavaScript
			__SEND_MESG_DEBUG("testClassACL after call in area '%s': class='%s' id_permission=%d \n", par_area, par_class, cnx->id_permission)
			__GET_RETURNED_INTEGER(acl)
			__CREATE_JS_FROM_INTEGER(acl)
			%%
			;
		};
	}; //end of method testClassACL

  method insertChild
  {
     doc "This method is overloaded of generic::insertChild. In this method we add copying of acl from parent";
     id_permission __PERMISSION_INSERT_CHILD;
     sec_type __METHOD_SEC_TYPE_OBJECT;
     
    __PARAMS_HEADER    
    param child_class 
         { 
           type CO_STRING; 
           doc "this is the name of class to insert record in";
         };
    param child_prop 
        { 
           type CO_STRING;
           doc "this is the property name in child class that holds id of parent";
        };
    param foreign_prop 
       { 
           type CO_STRING; 
           doc "this is the prop in container (this:) class that holds id to copy to child class";
       };
 
     declare language_c
     {
       C "int security_type;" ;
       C "int id_new;"; 
     };  
       
     body language_c
     {
       C
       %%
 
         // we switch off security for this one method! be carefull!
         cnx->check_security = 0;
         @co_call(par_area,'"generic"',"insertChild",par_area=>par_area,par_class=>par_class,par_id=>par_id,child_class=>child_class,child_prop=>child_prop, foreign_prop=>foreign_prop);
         id_new = *(int *) cou_getReturnValue(cnx, CO_INTEGER, &err);
         __SEND_MESG_DEBUG("obtained id=%d",id_new);
	 @co_call(par_area,par_class,"copyACL", \
	    par_area=>par_area,par_class=>par_class,par_id=>par_id, \
	    area_target=>par_area,class_target=>child_class,id_target=>id_new);
         cnx->check_security = 1;
         // security switched on again
/*
         __CREATE_QUERY("INSERT INTO co_sec_acl 
                         SELECT NEXTVAL('co_sec_acl_sqid'), version, deleted, id_permission, %d ,id_role, area, '%s'
                         FROM co_sec_acl
                         WHERE area = '%s' AND class_name = '%s' AND id_object = %d ",
                             id_new, child_class, par_area, par_class, par_id );
         __SQL_EXEC
*/
         cou_setReturnValue(cnx, &id_new, CO_INTEGER);         
       %%
       ;
     };
  };		   

	method insertChildX
	{
		doc "This method is a version of generic_acl::insertChild. In this method we set properties of the child.";
		id_permission __PERMISSION_INSERT_CHILD;
		sec_type __METHOD_SEC_TYPE_OBJECT;

     
		__PARAMS_HEADER    
		param child_class 
		{ 
			type CO_STRING; 
			doc "this is the name of class tu insert record in";
		};
		param child_prop 
		{ 
			type CO_STRING;
			doc "this is the property name in child class that holds id of parent";
		};
		param foreign_prop 
		{ 
			type CO_STRING; 
			doc "this is the prop in container (this:) class that holds id to copy to child class";
		};
		param create_params
		{
			type CO_ARRAY_OF_STRINGS;
			doc "this is array of additional parameters that we can pass to target class to create object; see create";
			optional 1;
			default "create_params = NULL;";
		};
 
 
		declare language_c
		{
			C "int security_type;" ;
			C "int id_new;"; 
			C 'int ppos;';
			C 'int NParams;';
		};  
       
		body language_c
		{
			C
			%%
 
			// we switch off security for this one method! be carefull!
			cnx->check_security = 0;
			@co_call(par_area,'"generic"',"insertChild", \
				par_area=>par_area,par_class=>par_class,par_id=>par_id, \
				child_class=>child_class,child_prop=>child_prop, \
				foreign_prop=>foreign_prop);
			id_new = *(int *) cou_getReturnValue(cnx, CO_INTEGER, &err);

			@co_call(par_area,par_class,"copyACL", \
				par_area=>par_area,par_class=>par_class,par_id=>par_id, \
				area_target=>par_area,class_target=>child_class,id_target=>id_new);
				cnx->check_security = 1;
         // security switched on again
/*
         __CREATE_QUERY("INSERT INTO co_sec_acl 
                         SELECT NEXTVAL('co_sec_acl_sqid'), version, deleted, id_permission, %d ,id_role, area, '%s'
                         FROM co_sec_acl
                         WHERE area = '%s' AND class_name = '%s' AND id_object = %d ",
                             id_new, child_class, par_area, par_class, par_id );
         __SQL_EXEC
*/
			if(create_params != NULL)
			{ 
				NParams = coa_GetElementCount(cnx, create_params);
				for(ppos = 0; ppos < NParams; ppos+=2)
				{
					const char* prop_name = coa_GetElement(cnx, create_params, ppos);
					const char* prop_value= coa_GetElement(cnx, create_params, ppos+1);
					@co_call(par_area, child_class, "update",par_area=>par_area,par_class=>child_class,par_id=>id_new,par_prop=>prop_name,new_value=>prop_value,version=>0);
				}
			}

			cou_setReturnValue(cnx, &id_new, CO_INTEGER);         
			%%
			;
		}; // body
	}; // method

  method cloneACL
  {
     doc "This method clones all ACLs from given object (used p.e. by catalog)";
     doc "You will want to overwrite this method if you have non-standard ACL checking.";
     id_permission __PERMISSION_CHANGE_ACL;
     sec_type __METHOD_SEC_TYPE_OBJECT;
     
    __PARAMS_HEADER    

    param area_parent 
         { 
           type CO_STRING; 
           doc "this is the name of the area of class that holds the ACLs";
         };
 
    param class_parent 
         { 
           type CO_STRING; 
           doc "this is the name of class that holds the ACLs";
         };
    param id_parent 
        { 
           type CO_INTEGER;
           doc "this is id of the element that holds the ACLs";
        };

     body language_c
     {
       C
       %%
         __CREATE_QUERY("INSERT INTO co_sec_acl 
                         SELECT NEXTVAL('co_sec_acl_sqid'), version, deleted, id_permission, %d ,id_role, '%s', '%s'
                         FROM co_sec_acl
                         WHERE area = '%s' AND class_name = '%s' AND id_object = %d ",
                         par_id, par_area, par_class, area_parent, class_parent, id_parent);
         __SQL_EXEC
       %%
       ;
     };
 };// end of cloneACL

  method copyACL
  {
     doc "This method copies all ACLs from object to the given object.";
     doc "Why not cloneACL? Because only class implementation know how to read ACLs.";
     doc "You will want to overwrite this method if you have non-standard ACL checking.";
     doc "This method is for use in methods only (permission=0; turn security off to use it).";
     id_permission 0;
     
    __PARAMS_HEADER    

    param area_target
         { 
           type CO_STRING; 
           doc "this is the name of the area of class to copy ACLs to.";
         };
 
    param class_target
         { 
           type CO_STRING; 
           doc "this is the name of class to copy ACLs to.";
         };
    param id_target
        { 
           type CO_INTEGER;
           doc "this is id of the element to copy ACLs to.";
        };

     body language_c
     {
       C
       %%
         __CREATE_QUERY("INSERT INTO co_sec_acl 
                         SELECT NEXTVAL('co_sec_acl_sqid'), version, deleted, id_permission, %d ,id_role, '%s', '%s'
                         FROM co_sec_acl
                         WHERE id_object = %d  AND class_name = '%s' AND area = '%s'",
                         id_target, area_target, class_target, par_id, par_class, par_area);
         __SQL_EXEC
       %%
       ;
     };
 };// end of copyACL

 method permissionOn
  {
     doc "This method sets one ACL On";
     id_permission __PERMISSION_CHANGE_ACL;
     sec_type __METHOD_SEC_TYPE_OBJECT;
     
    __PARAMS_HEADER_WITH_CALLBACK

    param id_permission 
         { 
           type CO_INTEGER; 
           doc "this is the id of permission that we wish to set on for current object";
         };
  
    param id_role
         { 
           type CO_INTEGER; 
			optional 1;
			default 'id_role = cnx->id_role;';
           doc "this is the id of the role that we want to grant permission";
         };
 
     body language_c
     {
       C
       %%
         // WE SHOULD CHECK WHETER WE CAN CONTROL THIS ROLE !!!!

         __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                         VALUES ( %d, %d, '%s', '%s', %d)",
                           id_permission, id_role, par_area, par_class, par_id);
 
         __SQL_EXEC
         __CREATE_JS_FROM_INTEGER(1)
 
       %%
       ;
     };
 }; // end of permissionOn

 method permissionOff
  {
     doc "This method sets one ACL On";
     id_permission __PERMISSION_CHANGE_ACL;
     sec_type __METHOD_SEC_TYPE_OBJECT;
     
    __PARAMS_HEADER_WITH_CALLBACK

    param id_permission 
         { 
           type CO_INTEGER; 
           doc "this is the id of permission that we wish to set on for current object";
         };
  
    param id_role
         { 
           type CO_INTEGER; 
           doc "this is the id of the role that we want to grant permission";
         };
 
     body language_c
     {
       C
       %%
         // WE SHOULD CHECK WHETER WE CAN CONTROL THIS ROLE !!!!

         __CREATE_QUERY("DELETE FROM co_sec_acl WHERE id_permission=%d AND id_role=%d AND area='%s' AND class_name='%s' AND id_object=%d",
                           id_permission, id_role, par_area, par_class, par_id);
 
         __SQL_EXEC
         __CREATE_JS_FROM_INTEGER(0)
       %%
       ;
     };
 }; // end of permissionOn


 method classPermissionOn
  {
     doc "This method sets one class ACL On";
     id_permission __PERMISSION_CHANGE_ACL;
     sec_type __METHOD_SEC_TYPE_CLASS;
     
    __PARAMS_HEADER_WITH_CALLBACK

    param id_permission 
         { 
           type CO_INTEGER; 
           doc "this is the id of permission that we wish to set on for current object";
         };
  
    param id_role
         { 
           type CO_INTEGER; 
			optional 1;
			default 'id_role = cnx->id_role;';
           doc "this is the id of the role that we want to grant permission";
         };
 
     body language_c
     {
       C
       %%
         // WE SHOULD CHECK WHETER WE CAN CONTROL THIS ROLE !!!!

         __CREATE_QUERY("INSERT INTO co_sec_acl_class (id_permission, id_role, area, class_name)
                         VALUES ( %d, %d, '%s', '%s')",
                           id_permission, id_role, par_area, par_class);
 
         __SQL_EXEC
         __CREATE_JS_FROM_INTEGER(1)
 
       %%
       ;
     };
 }; // end of classPermissionOn

 method classPermissionOff
  {
     doc "This method sets one class ACL Off";
     id_permission __PERMISSION_CHANGE_ACL;
     sec_type __METHOD_SEC_TYPE_CLASS;
     
    __PARAMS_HEADER_WITH_CALLBACK

    param id_permission 
         { 
           type CO_INTEGER; 
           doc "this is the id of permission that we wish to set on for current object";
         };
  
    param id_role
         { 
           type CO_INTEGER; 
           doc "this is the id of the role that we want to grant permission";
         };
 
     body language_c
     {
       C
       %%
         // WE SHOULD CHECK WHETER WE CAN CONTROL THIS ROLE !!!!

         __CREATE_QUERY("DELETE FROM co_sec_acl_class WHERE id_permission=%d AND id_role=%d AND area='%s' AND class_name='%s'",
                           id_permission, id_role, par_area, par_class);
 
         __SQL_EXEC
         __CREATE_JS_FROM_INTEGER(0)
       %%
       ;
     };
 }; // end of classPermissionOff


  method create
  {
     doc "this method creates empty object and sets ACL to allow changing of acls to active role";

     id_permission __PERMISSION_CREATE;
     sec_type __METHOD_SEC_TYPE_CLASS;

     __PARAMS_HEADER_WITH_CALLBACK

     param create_params
     {
         type CO_ARRAY_OF_STRINGS;
         optional 1;
	 default "create_params=NULL";
	 doc "this is array of additional parameters that we can pass to target class to create object";
     };
 
	param dont_set_acl
	{
		doc "If this param is set to 1 we do not set ACL for __PERMISSION_CHANGE_ACL for created object";
		type CO_INTEGER;
		optional 1;
		default "dont_set_acl = 0;";
	};

   declare language_c
    {
      C 
      %%
        int new_id=0;
	int i;
	int nFields;
        char* puta_classes[2]={NULL,NULL};
	int puta_ids[2]={0,0}; // for queueActionX 
	char **props;
	int check_security;
	int send_data;
	int id_role;
      %%
      ;	
    };  
 
     body language_c
     {
         C 
         %%
      __CREATE_QUERY("INSERT INTO co_%s_%s (deleted) VALUES (0)", 
                           par_area, par_class)

      __SQL_EXEC

      __SQL_GET_INTEGER(new_id, "SELECT last_value FROM co_%s_%s_sqid", par_area, par_class)

      puta_classes[0] = (char *) par_class;
      puta_ids[0] = new_id;
      
      cou_queueActionX( cnx, par_class, (const char **) puta_classes, puta_ids, CO_OP_CREATE, NULL);

      // getting current values

      __CREATE_QUERY("SELECT * FROM co_%s_%s WHERE id = %d",
                           par_area, par_class, new_id)

      __SQL_GET_RESULT
      
      puta_classes[0] = (char *) par_class;
      puta_ids[0] = new_id;

      nFields = cor_GetFieldCount(cnx, corec);
      props = ap_palloc(p, 2 * nFields * sizeof(char*) + 1);
      
      for (i = 0; i < nFields; i++)
      {
	 props[2*i]   = ap_pstrdup(p, cor_GetFieldName(cnx, corec, i));
	 props[2*i+1] = ap_pstrdup(p, cor_GetValue(cnx, corec, 0, i));
      } 
      props[2*i]=NULL;
      
      //cou_sendMessage(cnx, "we have props");
      
      __CLEAR_RESULT
      cou_queueActionXX( cnx, par_class, (const char **) puta_classes, puta_ids, CO_OP_UPDATE, (const char **) props);
      cou_setReturnValue(cnx, &new_id, CO_INTEGER);

      //setting ACL for current role
	if(!dont_set_acl)
	{	
		id_role = cnx->id_role;
		check_security = cnx->check_security;
        cnx->check_security = 0;
		send_data = cnx->send_data;
		cnx->send_data=0;
        @co_call(par_area,par_class,"permissionOn", \
            par_area=>par_area,par_class=>par_class,par_id=>new_id, \
            id_permission=>__PERMISSION_CHANGE_ACL,id_role=>id_role);
		cnx->send_data = send_data;
		cnx->check_security = check_security;
	}

     %%
     ;
     };
  }; // end of create

	method createAndSet
	{
		doc "this method creates empty object and sets ACL to allow changing of acls to active role, also sets props as given in create_params";

		id_permission __PERMISSION_CREATE;
		sec_type __METHOD_SEC_TYPE_CLASS;

		__PARAMS_HEADER_WITH_CALLBACK

		param create_params
		{
			type CO_ARRAY_OF_STRINGS;
			optional 1;
			default "create_params=NULL";
			doc "this is array of additional parameters that we can pass to target class to create object";
		};
 
		param dont_set_acl
		{
			doc "If this param is set to 1 we do not set ACL for __PERMISSION_CHANGE_ACL for created object";
			type CO_INTEGER;
			optional 1;
			default "dont_set_acl = 0;";
		};


		declare language_c
		{
			C 
			%%
			int new_id=0;
			int i;
			int nFields;
			int check_security;
			%%
			;	
		};  
 
		body language_c
		{
			C 
			%%

			//creating
			@co_lcall(par_area, par_class, "create", par_area=>par_area, par_class=>par_class, par_id=>par_id, create_params=>create_params, dont_set_acl=>dont_set_acl);
			__GET_RETURNED_INTEGER(new_id)
			__ASSERT(new_id>0, "Error during creating object")

			// setting props
			check_security = cnx->check_security;
			cnx->check_security = 0;

			nFields = coa_GetElementCount(cnx, create_params);
			for(i=0; i<nFields; i+=2)
			{
				__SEND_MESG_DEBUG("setting: %s", coa_GetStringElement(cnx, create_params, i))
				@co_call(par_area,par_class,"update", par_area=>par_area,par_class=>par_class,par_id=>new_id, par_prop=>'coa_GetStringElement(cnx, create_params, i)', new_value=>'coa_GetStringElement(cnx, create_params, i+1)', version=>-1);
			}
			cnx->check_security = check_security;
			%%
			;
		};
	}; // end of createAndSet

	method createAndSetWithACL
	{
		doc "this method creates object and sets ACLs taken from acl_class, also sets props as given in create_params";

		id_permission __PERMISSION_CREATE;
		sec_type __METHOD_SEC_TYPE_CLASS;

		__PARAMS_HEADER_WITH_CALLBACK

		param create_params
		{
			type CO_ARRAY_OF_STRINGS;
			optional 1;
			default "create_params=NULL";
			doc "this is array of additional parameters that we can pass to target class to create object";
		};
 
		declare language_c
		{
			C "int new_id=0;";	
		};  
 
		body language_c
		{
			C 
			%%

			//creating
			@co_lcall(par_area, par_class, "createAndSet", par_area=>par_area, par_class=>par_class, par_id=>par_id, create_params=>create_params, dont_set_acl=>1);
			__GET_RETURNED_INTEGER(new_id)
			__ASSERT(new_id>0, "Error during creating object")


			__CREATE_QUERY(" INSERT INTO co_sec_acl(id_permission, id_role, area, class_name, id_object) " 
                           " SELECT id_permission, id_role, area, class_name,  %d from co_sec_acl_class where class_name = '%s' and area='%s'",
			                 new_id, par_class, par_area)
			__SQL_EXEC						 
	
			%%
			;
		};
	}; // end of createAndSet





};  // end of class generic_acl
