#include "../../include/types.t"      
#include "../../include/defines.t"      

// CLASS MENU

class menu

{ 
  area "sys";
  area "sec";
  area "usr";

#include "../../include/check_acl.t"
  
  inherits catalog;
  
  container_class 'menu';
  container_column_name 'id';
      
 // PROPERTIES
      
#include "../../include/generic_acl_props.t"      

  prop name    
  {      
    type CO_STRING;      
    default "''";      
    params '';      
    representant 100;      
    foreign '';      
    foreign_prop '';      
    position 200;    
  };    
  
  prop id_parent
  {
    type CO_INTEGER;
    default 0;
    params '';
    representant 0;
    foreign 'menu';
    foreign_prop 'id';
    position 300;
  };
   
  prop id_type
  {
    type CO_STRING;
    default "''";
    params '';
    representant 0;
    foreign '';
    foreign_prop '';
    position 400;
  };  

  prop id_menu_action
  {
    type CO_INTEGER;
    default 0;
    params '';
    representant 0;
    foreign 'menu_action';
    foreign_prop 'id';
    position 500;
  };
  
  prop class_name
  {
    type CO_STRING;
    default "''";
    params '';
    representant 0;
    foreign 'meta_class';
    foreign_prop 'class_name';
    position 600;
  };  

  prop description
  {      
    type CO_STRING;      
    default "''";      
    params 'SIZE=100';      
    representant 0;      
    foreign '';      
    foreign_prop '';      
    position 700;    
  };    

  prop id_permission
  {      
    type CO_INTEGER;      
    default 0;      
    params '';      
    representant 0;      
    foreign '';      
    foreign_prop '';      
    position 800;    
  };    

  prop menu_position
  {      
    type CO_INTEGER;      
    default 0;      
    params '';      
    representant 0;      
    foreign '';      
    foreign_prop '';      
    position 900;    
  };

  prop display_name
  {    
    type CO_STRING;      
    default "''";      
    params '';      
    representant 0;      
    foreign '';      
    foreign_prop '';      
    position 1000;    
  };    

  prop menu_params
  {    
    type CO_STRING;      
    default "''";      
    params 'ROWS=10; COLS=30;';      
    representant 0;      
    foreign '';      
    foreign_prop '';      
    position 1200;    
  };    

   prop sec_type
   {
     type CO_INTEGER;
     default 0;
     params 'VALUES=new Array(0,1);';
     representant 0;
     foreign '';
     foreign_prop '';
     position 1300;
   };



include_c "../generic/co_generic.h";
include_c "../meta_class/co_meta_class.h";

  method getDefault
  {
   id_permission __PERMISSION_SELECT;
   sec_type __METHOD_SEC_TYPE_CLASS; 
   
    __PARAMS_HEADER_WITH_CALLBACK
 
   declare language_c
      {
        C "char * menu_name;";
        C "int id_menu;";
      };


    body language_c
      {
         C
	 %%
	 cou_sendMessageX(cnx, "entering getDefault", CO_MSG_DEBUG);
         
         menu_name = ap_psprintf(p, "default_%s", par_class);
         @co_getPropByName(par_area, '"menu"', menu_name, '"id"');
         if(err<0)
         {
            cou_sendMessageX(cnx, "no default menu for this class", CO_MSG_ERROR);
            return -1;
         }
         id_menu = atoi(get_str);
         @co_getRelated(par_area, '"menu"', id_menu, '"menu"');
         __CREATE_JS
	 cou_sendMessageX(cnx, "leaving getDefault", CO_MSG_DEBUG);
	 %%
	 ;
      };  
								    
  };

  method getByName
  {
   id_permission __PERMISSION_SELECT;
   sec_type __METHOD_SEC_TYPE_CLASS; 

   
    __PARAMS_HEADER_WITH_CALLBACK

   param menu_name {type CO_STRING;};
 
   declare language_c
      {
        C "int id_menu;";
      };


    body language_c
      {
         C
	 %%
//
         @co_getPropByName(par_area, '"menu"', menu_name, '"id"');
         if(err<0)
         {
            cou_sendMessageX(cnx, "no menu for this name", CO_MSG_ERROR);
            return -1;
         }
         id_menu = atoi(get_str);
         @co_getRelated(par_area, '"menu"', id_menu, '"menu"');
         __CREATE_JS
	 %%
	 ;
      };  
								    
 };
 
  method getByNameAllowed
  {
   id_permission __PERMISSION_SELECT;
   sec_type __METHOD_SEC_TYPE_CLASS; 

   
   doc 
      %%
        this method retrieves menu items, that are allowed to execute for current users role and given object
        Objects are identified using: for_area::for_class::for_id
      %%
      ;

    __PARAMS_HEADER_WITH_CALLBACK

   param menu_name 
   {
     doc "name for menu to retrieve";
     type CO_STRING;
   };
   param for_area 
   {
     doc "area for object that we use to check permissions";
     type CO_STRING;
   };
   param for_class 
   {
     doc "class for object that we use to check permissions";
     type CO_STRING;
   };
   param for_id 
   {
     doc "id for object that we use to check permissions";
     type CO_INTEGER;
   };

   declare language_c
      {
        C "int id_menu;";
      };


    body language_c
      {
         C
	 %%
//
         @co_getPropByName(par_area, '"menu"', menu_name, '"id"');
         if(err<0)
         {
            cou_sendMessageX(cnx, "no menu for this name", CO_MSG_ERROR);
            return -1;
         }
         id_menu = atoi(get_str);
 
         __CREATE_QUERY(" SELECT DISTINCT m.* "
                        " FROM co_%s_menu m, co_sec_acl a "
                        " WHERE m.id_parent = %d "
                        " AND a.id_permission = m.id_permission AND a.area = '%s' AND a.class_name = '%s' AND a.id_object = %d "
                        " AND a.id_role = %d "
                        " ORDER BY menu_position "   ,
                        par_area, id_menu, for_area, for_class, for_id, cnx->id_role
                       )
         __GET_RESULT_AND_CREATE_JS
        %%
	 ;
      }; // body
 };//method


	method getByNameAllowedWithParents
	{

		id_permission __PERMISSION_SELECT;
		sec_type __METHOD_SEC_TYPE_CLASS; 
   
		doc 
		%%
		This method retrieves menu items, that are allowed to execute for current users role and given object
		Objects are identified using: for_area::for_class::for_id
		This method retrieves menus for given class and all inheritance path
		Name of menu is created by adding '#type@class_name' at the end, so if you want to retrieve menu 'default#single@catalog' 
		as name you provide only 'default', and type 'single'
		You will get items from menus: 'default#single@catalog', 'default#single@generic_acl', 'default#single@generic'
		%%
		{
			pl
			%%
			Ta metoda zwraca elementy menu, ktore odpowiadaja akcjom, ktore moze wykonac aktywny uzytkownik
			Obiekty sa identyfikowane przez: for_area::for_class::for_id
			Ta metoda zwraca menu dla podanej klasy i dla calej sciezki dziedziczenia
			Do nazwy menu jest dodawana na koncu typ i nazwa klasy '#type@class_name', jesli wiec chcesz uzyskac menu 'default#single@catalog'
			jako nazwe podajesz tylko 'default' i jako type 'single'
			Uzyskasz elementy dla menu:  'default#single@catalog', 'default#single@generic_acl', 'default#single@generic'
			%%
			;
		}; // doc


		__PARAMS_HEADER_WITH_CALLBACK
	
		param menu_name 
		{
			doc "name for menu to retrieve";
			type CO_STRING;
		};
		param for_area 
		{
			doc "area for object that we use to check permissions";
			type CO_STRING;
		};
		param for_class 
		{
			doc "class for object that we use to check permissions";
			type CO_STRING;
		};
		param for_id 
		{
			doc "Id for object that we use to check permissions. Default value is -1, so you do not have to pass this for multi and class methods";
			optional 1;
			default 'for_id = -1;';
			type CO_INTEGER;
		};

		param type
		{
			doc "Type of catalog to be retrieved (like single, class, multi)";
			type CO_STRING;
			optional 1;
			default 'type = ap_psprintf(p,"single");';
		};

		declare language_c
		{
			C "int id_menu;";
			C 'char * path=NULL;';
			C 'char * class_names[1000];';
			C 'const char * cur_menu_name = "";';
			C 'int i=0;';
		};


		body language_c
		{
			C
			%%
			memset(class_names, 0, sizeof(char*)*1000);

			@co_lcall(par_area,'"meta_class"',"getFullPathByClassName",par_area=>par_area,par_class=>'"meta_class"',par_id=>-1,callback=>'""',class_name=>for_class);
			__GET_RETURNED_STRING(path)

			/* WARNING: MODIFIES READ ReturnValue ! */
			cou_split(path,class_names,"|",1000);
          
			i=0;
			while(class_names[i])
			{
				cur_menu_name = ap_psprintf(p, "%s#%s@%s", menu_name, type, class_names[i]);
				__SEND_MESG_DEBUG(cur_menu_name)
				@co_getPropByName(par_area, '"menu"', cur_menu_name, '"id"');
				if(err==0)
				{
					id_menu = atoi(get_str);

					if(strlen(SQLString)>0) __APPEND_TO_QUERY(" UNION \n");
 
					__APPEND_TO_QUERY(
							" SELECT DISTINCT m.* "
							" FROM co_%s_menu m "
							" WHERE m.id_parent = %d  AND "
							" CASE WHEN sec_type=1 THEN co_%s_%s_check_acl_class(id_permission, %d) " 
							" ELSE co_%s_%s_check_acl(%d, id_permission, %d) END = 1"
							,par_area, id_menu, 
							for_area, for_class, cnx->id_role,
							for_area, for_class, for_id, cnx->id_role
							)
				}// if
				i++;
			} // while
			//__ASSERT(strlen(SQLString)>0, "Menu not found")
			if (strlen(SQLString)>0) // something was found - we have query
			{
				__APPEND_TO_QUERY(" ORDER BY menu_position ")
				__GET_RESULT_AND_CREATE_JS
			}
	       	%%
			;
		};// body
	}; //method

}; //class
 


//END OF CLASS MENU
