// TYPES FOR CO

#include "../../include/types.t"
#include "../../include/defines.t"


// CLASS spreadsheet_cevent

class spreadsheet_cevent
{
  area sys;

#include "check_iacl.t"

 // PROPERTIES

#include "generic_props.t"

  inherits generic_iacl;

  uses "js_code";
  uses "generic";
  uses "generic_acl";

  container_class 'spreadsheet';
  container_column_name 'spreadsheet_id';

  prop spreadsheet_id
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'spreadsheet';
   foreign_prop 'id';
   position 44;
  };

  prop column_id
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'spreadsheet_column';
   foreign_prop 'id';
   position 44;
  };

  prop row_id
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'spreadsheet_row';
   foreign_prop 'id';
   position 44;
  };

  prop event_type_id
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'cevent_type';
   foreign_prop 'id';
   position 44;
  };

  prop id_js_code
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'js_code';
   foreign_prop 'id';
   position 44;
  };

 // METHODS

  method fire
  {
    doc "for future use (or removal)";
    id_permission __PERMISSION_EVALUATE;
    sec_type __METHOD_SEC_TYPE_OBJECT;

    __PARAMS_HEADER
    body language_c
    {
     C
     %%
	/* get context info, create prolog (context.id_spreadsheet etc.), use
	    js_code.evaluate to fire event */

     %%
     ;
    };
  };

    method getObjectEvents
    {
      doc "returns recordset with events attachet to given(spreadsheet part) object";
      id_permission __PERMISSION_GET;
      sec_type __METHOD_SEC_TYPE_OBJECT;

      __PARAMS_HEADER

      param id_spreadsheet { type CO_INTEGER; };
      param id_column { type CO_INTEGER; };
      param id_row { type CO_INTEGER; };

      body language_c
      {
        C
        %%
        // here we should gather events info
        __CREATE_QUERY(
	"SELECT ET.type AS cevtype, SCR.code AS code, E.id_js_code AS id_js_code, E.id AS id, SCR.version AS version "
        "FROM co_%s_%s E, co_%s_cevent_type ET, co_%s_js_code SCR "
        "WHERE (E.event_type_id=ET.id) AND "
	"(E.spreadsheet_id=%d) AND (E.column_id=%d) AND (E.row_id=%d) AND (SCR.id=E.id_js_code)",
	par_area, par_class, par_area, par_area, 
	id_spreadsheet, id_column, id_row)

        __SQL_GET_RESULT

        __RETURN_RESULT

        %%
        ;
      };
    };

  method setEventScript
  {
    doc "rather obsolete, can call update method directly";
    id_permission __PERMISSION_UPDATE;
    sec_type __METHOD_SEC_TYPE_OBJECT;

    __PARAMS_HEADER
    param event_script { type CO_STRING; };
    param version{ type CO_INTEGER; };

    declare language_c
    {
    C
    %%
    int id_js_code;
    %%
    ;
    };

    body language_c
    {
    C
    %%
//
        @co_getProp(par_area, par_class, par_id,'"id_js_code"');

	if(err==0)
	{
	    id_js_code = atoi(get_str);
	}
	else
	{
	    cou_sendMessage(cnx, ap_psprintf(p, "could not find event info for id: %d",par_id));
	    return -1;
	};

	@co_call(par_area, '"js_code"', "update", par_area=>par_area, par_class=>'"js_code"', par_id=>id_js_code, par_prop=>'"code"', new_value=>event_script ,version=>version);
    %%
    ;
    };
  };

  method createEventScript
  {
    doc "attach new event to spreadsheet (part) object";
    id_permission __PERMISSION_INSERT_CHILD;
    sec_type __METHOD_SEC_TYPE_CLASS;

    __PARAMS_HEADER
    param id_spreadsheet { type CO_INTEGER; };
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param event_type_id { type CO_INTEGER; };
    param event_script { type CO_STRING; };

    declare language_c
    {
    C
    %%
	int id_js_code;
	int id_cevent;
	int *ret_ptr;
	int my_err=0;
    %%
    ;
    };

    body language_c
    {
    C
    %%
//
    @co_call(par_area, '"js_code"', "createScript", par_area=>par_area, par_class=>'"js_code"', par_id=>-1, code=>event_script);
    __GET_RETURNED_INTEGER(id_js_code)
    if (!id_js_code) cou_error( cnx, "js_code::createScript returned 0");

    __SQL_GET_INTEGER( id_cevent, "
    INSERT INTO co_%s_%s (spreadsheet_id, column_id, row_id, event_type_id, id_js_code)
    VALUES(%d, %d, %d, %d, %d);
    SELECT last_value FROM co_%s_%s_sqid;
    ",      par_area, par_class, 
	    id_spreadsheet, id_column, id_row, event_type_id, id_js_code,
    	    par_area, par_class )

    if (!(id_cevent>0)) cou_error( cnx, "Error while inserting spreadsheet cevent");

    cou_setReturnValue( cnx, &id_cevent, CO_INTEGER );
    cou_logError( cnx, CO_LOG_DEBUG, "Result: %d", id_cevent );

    {
        const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
        int object_ids[] = {id_spreadsheet,id_column,id_row};

        cou_queueActionX( cnx, "spreadsheet_cevent", class_names, object_ids,
                CO_OP_INSERT, 
		"event_type_id", ap_psprintf(p, "%d", event_type_id), 
		"id_js_code", ap_psprintf(p, "%d", id_js_code),
		"id", ap_psprintf(p, "%d", id_cevent),
		NULL );
    }
    %%
    ;
    };
  };

};

//END OF CLASS spreadsheet_cevent