 
          DROP FUNCTION co_sys_expression_range_getNVal(int,int,int,int)
 
          !NOERROR 
          CREATE FUNCTION  co_sys_expression_range_getNVal(int,int,int,int)

          RETURNS  float4
          AS ' 
 
DECLARE
    _id_spreadsheet ALIAS FOR $1;
    _id_column ALIAS FOR $2;
    _id_row ALIAS FOR $3;
    _index ALIAS FOR $4;
    _reference co_sys_expression_range%ROWTYPE;
    _coordinates RECORD;
    _coordinates2 RECORD;
    _result float4;
BEGIN
    SELECT * INTO _reference FROM co_sys_expression_range
        WHERE
        id_spreadsheet = _id_spreadsheet AND
        id_column = _id_column AND
        id_row = _id_row AND
        index_ref = _index
    ;
    SELECT COL.column_number, ROW.row_number INTO _coordinates
    FROM co_sys_spreadsheet_column AS COL,co_sys_spreadsheet_row AS ROW
        WHERE
        COL.id = _reference.id_ref_column AND
        ROW.id = _reference.id_ref_row
    ;
    -- for sum only
    IF NOT FOUND THEN
	RETURN 0;
    END IF;

    SELECT COL.column_number, ROW.row_number INTO _coordinates2
    FROM co_sys_spreadsheet_column AS COL,co_sys_spreadsheet_row AS ROW
        WHERE
        COL.id = _reference.id_ref_column2 AND
        ROW.id = _reference.id_ref_row2
    ;
    -- for sum only
    IF NOT FOUND THEN
	RETURN 0;
    END IF;

    SELECT SUM(CELL.value_num) INTO _result
    FROM co_sys_spreadsheet_cell AS CELL,
        co_sys_spreadsheet_column AS COL,
        co_sys_spreadsheet_row AS ROW
    WHERE
        COL.spreadsheet_id = _reference.id_ref_spreadsheet AND
        COL.column_number >= _coordinates.column_number AND
        COL.column_number <= _coordinates2.column_number AND
        ROW.spreadsheet_id = _reference.id_ref_spreadsheet AND
        ROW.row_number >= _coordinates.row_number AND
        ROW.row_number <= _coordinates2.row_number AND
        CELL.column_id = COL.id AND
        CELL.row_id = ROW.id
    ;
RETURN _result;

END;
    

    ' LANGUAGE 'plpgsql'
    !GO 
  
          DROP FUNCTION co_sys_expression_range_replRow(int,int,int,int)
 
          !NOERROR 
          CREATE FUNCTION  co_sys_expression_range_replRow(int,int,int,int)

          RETURNS  int
          AS ' 
 
DECLARE
    _id_spreadsheet ALIAS FOR $1;
    _id_row ALIAS FOR $2;
    _id_prev_row ALIAS FOR $3;
    _id_next_row ALIAS FOR $4;
BEGIN
    UPDATE co_sys_expression_range
    SET id_ref_row = _id_next_row
        WHERE
        id_spreadsheet = _id_spreadsheet AND
	id_ref_row = _id_row
    ;

    UPDATE co_sys_expression_range
    SET id_ref_row2 = _id_prev_row
        WHERE
        id_spreadsheet = _id_spreadsheet AND
	id_ref_row2 = _id_row
    ;
RETURN 0;

END;
    

    ' LANGUAGE 'plpgsql'
    !GO 
  
          DROP FUNCTION co_sys_expression_range_fillV(int,int,int,int)
 
          !NOERROR 
          CREATE FUNCTION  co_sys_expression_range_fillV(int,int,int,int)

          RETURNS  float4
          AS ' 
 
DECLARE
    _id_spreadsheet ALIAS FOR $1;
    _id_column ALIAS FOR $2;
    _id_row ALIAS FOR $3;
    _id_row_last ALIAS FOR $4;
    _nr_row_first INTEGER;
    _nr_row_last INTEGER;
    _reference co_sys_expression_range%ROWTYPE;
BEGIN
    SELECT row_number INTO _nr_row_first FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row
    ;
    SELECT row_number INTO _nr_row_last FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row_last
    ;
    DELETE FROM co_sys_expression_range
        WHERE EXISTS(
            SELECT * FROM co_sys_spreadsheet_row  AS ROW
                WHERE
                ROW.id = id_row AND
                ROW.row_number > _nr_row_first AND
                ROW.row_number <= _nr_row_last
            )
    ;
    /* referred column id never change; ref. row id may change */
    INSERT INTO co_sys_expression_range
        SELECT
        RANGE.version,
        RANGE.deleted,
        RANGE.id_spreadsheet,
        RANGE.id_column,
        ROW.id,
        RANGE.index_ref,
        RANGE.function,
        RANGE.id_ref_spreadsheet,
        NEWREFROW.id,/* update as you go; depends on is_ref_row_hard */
        RANGE.id_ref_column,
        RANGE.is_ref_row_hard,
        RANGE.is_ref_column_hard,
        RANGE.is_ref_row_named,
        RANGE.is_ref_column_named,
        NEWREFROW2.id,/* update as you go; depends on is_ref_row2_hard */
        RANGE.id_ref_column2,
        RANGE.is_ref_row2_hard,
        RANGE.is_ref_column2_hard,
        RANGE.is_ref_row2_named,
        RANGE.is_ref_column2_named

        FROM
            co_sys_expression_range AS RANGE,
            co_sys_spreadsheet_row AS ROW,
            co_sys_spreadsheet_row AS REFROW,
            co_sys_spreadsheet_row AS REFROW2,
            co_sys_spreadsheet_row AS NEWREFROW,
            co_sys_spreadsheet_row AS NEWREFROW2
        WHERE
            RANGE.id_spreadsheet = _id_spreadsheet AND
            RANGE.id_column = _id_column AND
            RANGE.id_row = _id_row AND
            ROW.spreadsheet_id = _id_spreadsheet AND
            ROW.row_number > _nr_row_first AND
            ROW.row_number <= _nr_row_last AND
            REFROW.id = RANGE.id_ref_row AND
            REFROW2.id = RANGE.id_ref_row2 AND
            NEWREFROW.spreadsheet_id = _id_spreadsheet AND
            (
            (NEWREFROW.row_number = REFROW.row_number + (ROW.row_number - _nr_row_first)
                AND NOT RANGE.is_ref_row_hard) OR (
                NEWREFROW.id = REFROW.id AND RANGE.is_ref_row_hard)
            ) AND
            NEWREFROW2.spreadsheet_id = _id_spreadsheet AND
            (
            (NEWREFROW2.row_number = REFROW2.row_number + (ROW.row_number - _nr_row_first)
                AND NOT RANGE.is_ref_row2_hard) OR (
                NEWREFROW2.id = REFROW2.id AND RANGE.is_ref_row2_hard)
            )
    ;
RETURN 0;

END;
    

    ' LANGUAGE 'plpgsql'
    !GO 
  
  DROP FUNCTION co_sys_expression_range_check_acl(int,int,int) 
  !NOERROR 
  CREATE FUNCTION  co_sys_expression_range_check_acl(int,int,int)
  RETURNS  int
  AS ' 
SELECT 1 FROM co_sec_acl WHERE
         area = \'sys\'
	 AND class_name = \'spreadsheet\'
	 AND id_object = $1
	 AND id_permission = $2
	 AND id_role = $3
	
' LANGUAGE 'sql' 
!GO 
  
  DROP FUNCTION co_sys_expression_range_check_acl_class(int,int)
  !NOERROR 
  CREATE FUNCTION  co_sys_expression_range_check_acl_class(int,int)
  RETURNS  int
  AS ' 
SELECT 1 FROM co_sec_acl_class WHERE
         area = \'sys\'
	 AND class_name = \'spreadsheet\'
	 AND id_permission = $1
	 AND id_role = $2
	
 
 ' LANGUAGE 'sql'   
    !GO 
  
     DROP FUNCTION lhlo_sys_expression_range_version(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_version(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.version = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_deleted(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_deleted(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.deleted = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_spreadsheet(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_spreadsheet(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_spreadsheet = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_column(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_column(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_column = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_row(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_row(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_row = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_index_ref(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_index_ref(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.index_ref = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_function(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_function(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.function = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_ref_spreadsheet(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_ref_spreadsheet(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_ref_spreadsheet = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_ref_row(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_ref_row(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_ref_row = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_ref_column(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_ref_column(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_ref_column = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_row_hard(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_row_hard(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_row_hard = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_column_hard(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_column_hard(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_column_hard = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_row_named(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_row_named(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_row_named = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_column_named(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_column_named(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_column_named = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_ref_row2(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_ref_row2(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_ref_row2 = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_id_ref_column2(int)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_id_ref_column2(int) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.id_ref_column2 = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_row2_hard(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_row2_hard(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_row2_hard = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_column2_hard(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_column2_hard(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_column2_hard = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_row2_named(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_row2_named(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_row2_named = $1
     '
     LANGUAGE 'sql'
     !GO
        
     DROP FUNCTION lhlo_sys_expression_range_is_ref_column2_named(bool)
     !NOERROR

     CREATE FUNCTION lhlo_sys_expression_range_is_ref_column2_named(bool) RETURNS co_sys_expression_range AS
     '
        SELECT * FROM co_sys_expression_range WHERE co_sys_expression_range.is_ref_column2_named = $1
     '
     LANGUAGE 'sql'
     !GO
       