###################################################################################
# SUB: co_get_param
# IN: param_name, co_type, optional, default_val
# OUT: string containing code for initializing param from request
# It is for "static" use only at the begining of methos. You could not use this
# macro for dynamic reading of params inside method.
###################################################################################

sub co_get_param
{
  my ($param_name, $co_type, $optional, $default_val, $name_class, $name_method, $return_null) = @_;
  my ($ret_type) = $co_type_map{$co_type}{"c_type"};
  my ($pointer) =  $co_type_map{$co_type}{"pointer"};
  my ($is_array) = $co_type_map{$co_type}{"is_array"};

  my ($ret);
  if($pointer)
  { 
    if($is_array)
    {
      $ret = "$param_name = ($ret_type)cou_getParamArray(cnx,\"_$param_name\",delim,MAXSELECTPAR, &err);\n"; 
    }
    else
    {  
      $ret = "$param_name = ($ret_type)cou_getParam(cnx,\"$param_name\",$co_type, &err);\n"; 
    }  
  }
  else
  { 
    $ret = "{ void *ret=cou_getParam(cnx,\"$param_name\",$co_type, &err);
              if(ret != NULL)
	      {
                ${param_name} = *($ret_type *)ret;
	      }
	      else
	      {
                ${param_name} = ($ret_type)0;
	      }
	    }\n"; 
  }
  
  if($optional)
  {
    $ret .= "if(err==CO_PARAMETER_NOT_FOUND){ ${default_val}; err=0; }\n";
  }
  if($return_null)
  {
    $ret .= "__ASSERT_NULL(err>=0, \"in ${name_class}.${name_method} parameter not optional: $param_name\")\n";
  }
  else
  {
    $ret .= "__ASSERT(err>=0, \"in ${name_class}.${name_method} parameter not optional: $param_name\")\n";
  }
  return($ret);
}  

###################################################################################
# SUB: co_call
# IN: area, class, met, params
# OUT: string containing code for calling method of class using cou_broker
# Typical usage is like this:
# co_call(par_area, par_class, '"get"', par_area=>par_area, par_class=>par_class, par_id=>par_id);
# You can not pass method name as name of variable it HAS TO BE static string.
# Becouse we need that during compilation!
# This mathod is not changing jsPut, so called method is puting JavaScript to client
####################################################################################

sub co_call
{
  my ($area, $class, $met, %params)= @_;
  
  my ($ret);
  $co_call_number++;
  
  my ($param_name) = "out_param".$co_call_number;
  $ret = "
  {
   __PARAM_STRUCT_CREATE_${met}(${param_name})
  ";
  foreach $param (keys %params)
  {
    $ret .= "__PARAM_STRUCT_SET($params{$param} , ${param_name}, ${param})\n";
  };
  $ret .= "cou_broker($area,$class,\"$met\",cnx,out_param$co_call_number);\n}\n";
  return($ret);
}      

###################################################################################
# SUB: co_lcall
# IN: area, class, met, params
# OUT: string containing code for calling method of class using cou_broker
# Same as co_call (co_lcall means local call) but: 
# This mathod sets jsPut to false, so called method is NOT puting JavaScript to client
####################################################################################
sub co_lcall
{
  my ($area, $class, $met, %params)= @_;
  
  my ($ret);
  $co_call_number++;
  
  my ($param_name) = "out_param".$co_call_number;
  $ret  = "{  
             __PARAM_STRUCT_CREATE_${met}(${param_name})
            co_client_mode old_client_mode = cnx->client_mode;
            cnx->client_mode=CO_CLIENT_NO_MODE;
  ";

  foreach $param (keys %params)
  {
    $ret .= "__PARAM_STRUCT_SET($params{$param} , ${param_name}, ${param})\n";
  };
  $ret .= "cou_broker($area,$class,\"$met\",cnx,out_param$co_call_number);\n";
  $ret .= "cnx->client_mode = old_client_mode;\n";
  $ret .= "}\n";
  return($ret);
}      

###################################################################################
# SUB: co_getProp
# IN: area, class, id, prop
# OUT: string containing code for calling method to get one property from one record
# Typical usage is like this:
# co_getProp(par_area, par_class, some_int, prop_name);
# the value of property is written to get_str variable
# get_str is set to NULL if property not found
# This mathod sets jsPut to false, so called method is NOT puting JavaScript to client
####################################################################################
  
sub co_getProp
{
  my ($area, $class, $id, $prop)= @_;
  $co_call_number++;
  my ($param_name) = "out_param".$co_call_number;
  my ($met) = "get";
  my(%params) = ( par_area => $area, par_class => $class, par_id => $id );
  
  my($ret)  = "
  err=0;
  {
     __PARAM_STRUCT_CREATE_${met}(${param_name})
     co_recordset * loc_corec;
     co_client_mode old_client_mode = cnx->client_mode;
     cnx->client_mode=CO_CLIENT_NO_MODE;
  ";
    
  foreach $param (keys %params)
  {
    $ret .= "__PARAM_STRUCT_SET($params{$param} , ${param_name}, ${param})\n";
  };
  $ret .= "
   cou_broker($area,$class,\"$met\",cnx,$param_name);
   cnx->client_mode = old_client_mode;
   loc_corec = cou_getReturnValue(cnx, CO_RECORDSET, &err);
   if(cor_GetTupleCount(cnx, loc_corec)>0)
   {
      get_str = cor_GetCopyOfValueByName(cnx, loc_corec,0,$prop);
   }
   else
   {
      err=-1;
      get_str = NULL;
   };            
  }";
  return($ret);
}

###################################################################################
# SUB: co_getPropByName
# IN: area, class, name, prop
# OUT: string containing code for calling method to get one property from one record
# Typical usage is like this:
# co_getPropByName(par_area, par_class, name_to_find_string, prop_name);
# Same as co_getProp but instead of id it uses name prop to locate record
# IT DOES NOT CHECK WHETHER WE FOUND SOMETHING -- will be fixed soon
# get_str is set to NULL if property not found
# This method sets jsPut to false, so called method is NOT puting JavaScript to client
####################################################################################

sub co_getPropByName
{
  my ($area, $class, $name, $prop)= @_;
  $co_call_number++;
  my ($param_name) = "out_param".$co_call_number;
  my ($met) = "getByProp";
  my(%params) = ( par_area => $area, par_class => $class, par_id => -1, prop_name=>'"name"', prop_val=>$name );
  
  my($ret)  = "
  err=0;
  {
           __PARAM_STRUCT_CREATE_${met}(${param_name})
           co_recordset * loc_corec;
           co_client_mode old_client_mode = cnx->client_mode;
  	   cnx->client_mode=CO_CLIENT_NO_MODE;
           cou_sendMessageX(cnx, \"param struct created\", CO_MSG_DEBUG); 

 ";

  foreach $param (keys %params)
  {
    $ret .= "__PARAM_STRUCT_SET($params{$param} , ${param_name}, ${param})\n";
  };
  $ret .= "
          cou_sendMessageX(cnx, \"param struct values set\", CO_MSG_DEBUG); 

           cou_broker($area,$class,\"$met\",cnx,$param_name);
           cnx->client_mode = old_client_mode;
           loc_corec = cou_getReturnValue(cnx, CO_RECORDSET, &err);
	   if(cor_GetTupleCount(cnx, loc_corec)>0)
           {
	      get_str = cor_GetCopyOfValueByName(cnx, loc_corec,0,$prop);
           }
           else
           {
              err=-1;
              get_str = NULL;
           };
       }\n";

  return($ret);
}

###################################################################################
# SUB: co_getRelated
# IN: area, class, id, related_class
# OUT: string containing code for calling method to get recordset with related records
# Typical usage is like this:
# co_getRelated(par_area, par_class, some_int, class_name);
# This is for relation like company -> employe 
# relation has to described in meta_prop 
# This method sets jsPut to false, so called method is NOT puting JavaScript to client
####################################################################################

sub co_getRelated
{
  my ($area, $class, $id, $related_class)= @_;
  $co_call_number++;
  my ($param_name) = "out_param".$co_call_number;
  my ($met) = "getRelatedData";
  
  my(%params) = ( par_area => $area, par_class => $class, par_id => $id, related_class => $related_class );
  
  $ret  = "err=0;
  {
           __PARAM_STRUCT_CREATE_${met}(${param_name})
            co_client_mode old_client_mode = cnx->client_mode;
	    cnx->client_mode=CO_CLIENT_NO_MODE;
  ";
	    
  foreach $param (keys %params)
  {
    $ret .= "__PARAM_STRUCT_SET($params{$param} , ${param_name}, ${param})\n";
  };
  $ret .= "
           cou_broker($area,$class,\"$met\",cnx,$param_name);
           cnx->client_mode = old_client_mode;
           corec = cou_getReturnValue(cnx, CO_RECORDSET, &err);
           }\n";
  return($ret);
}

###################################################################################
# SUB: co_filterByProp
# IN: area, class, prop_name, filter_name, filter_value
# OUT: string containing code for calling method to select like:
# SELECT class.prop_name FROM class WHERE class.filter_name LIKE filter_value
# Typical usage is like this:
# co_getProp(par_area, par_class, prop_to_read, prop_to_filter, filter_value);
# This method sets jsPut to false, so called method is NOT puting JavaScript to client
####################################################################################

sub co_filterByProp
{
  my ($area, $class, $prop_name, $filter_name, $filter_value)= @_;
  $co_call_number++;
  my ($param_name) = "out_param".$co_call_number;
  my ($met) = "selectX";
  
  my($ret)  = "
  err=0;
  {
    __PARAM_STRUCT_CREATE_${met}(${param_name})
    co_client_mode old_client_mode = cnx->client_mode;
    co_array_of_pointers * get_class_names       = coa_Create(cnx);
    co_array_of_pointers * get_prop_names        = coa_Create(cnx);
    co_array_of_pointers * filter_class_names    = coa_Create(cnx);
    co_array_of_pointers * filter_prop_names     = coa_Create(cnx);
    co_array_of_pointers * filter_operators      = coa_Create(cnx);
    co_array_of_pointers * filter_values         = coa_Create(cnx);
 
    coa_AppendElement(cnx, get_class_names, $class);
    coa_AppendElement(cnx, get_prop_names,  $prop_name);
    coa_AppendElement(cnx, filter_class_names, $class);
    coa_AppendElement(cnx, filter_prop_names, $filter_name);
    coa_AppendElement(cnx, filter_operators, \"LIKE\");
    coa_AppendElement(cnx, filter_values, $filter_value);  
    
    
    cnx->client_mode=CO_CLIENT_NO_MODE;

    __PARAM_STRUCT_SET($area,             ${param_name}, par_area)
    __PARAM_STRUCT_SET($class,            ${param_name}, par_class)
    __PARAM_STRUCT_SET(-1,                ${param_name}, par_id)
    __PARAM_STRUCT_SET(get_class_names,   ${param_name}, get_class_names)
    __PARAM_STRUCT_SET(get_prop_names,    ${param_name}, get_prop_names)
    __PARAM_STRUCT_SET(filter_class_names,${param_name}, filter_class_names)
    __PARAM_STRUCT_SET(filter_prop_names, ${param_name}, filter_prop_names)
    __PARAM_STRUCT_SET(filter_operators,  ${param_name}, filter_operators)
    __PARAM_STRUCT_SET(filter_values,     ${param_name}, filter_values)
 
    cou_broker($area,$class,\"$met\",cnx,$param_name);
    cnx->client_mode = old_client_mode;
    corec = cou_getReturnValue(cnx, CO_RECORDSET, &err);
  }
  ";
  return($ret);
}
 
1;
